r"""Helper classes for tree spaces.

Class ``Split``.
Essentially, a ``Split`` is a two-set partition of a subset of :math:`\{0,\dots,n-1\}`.
This class is designed such that one part of both parts of the partition can be empty.
Splits are corresponding uniquely to edges in a phylogenetic forest, where, if one cuts
the edge in the forest, the resulting two-set partition of the labels of the respective
component of the forest is the corresponding split.


Lead author: Jonas Lueg
"""

import functools
import itertools
import random

import geomstats.backend as gs
from geomstats.exceptions import NotPartialOrder


def _pop_random_elem(ls):
    """Pops a random element from a list.

    Parameters
    ----------
    ls : list
    """
    random_index = random.randint(0, len(ls) - 1)
    return ls.pop(random_index)


def generate_splits(labels):
    """Generate random maximal set of compatible splits of set ``labels``.

    This method works inductively on the number of elements in labels.
    Start with a split of two randomly chosen labels. Then, successively choose
    a label from the labels and add this as a leaf with a split to the existing
    tree by attaching it to a random split, thereby dividing this split into two
    splits and one has to update all the other splits accordingly.

    Parameters
    ----------
    labels : list[int]
        A list of integers, the set of labels that we generate splits for.

    Returns
    -------
    splits : list[Split]
        A list of splits of the set of labels, maximal number of splits.
    """
    if len(labels) <= 1:
        return []

    unused_labels = labels.copy()

    u = _pop_random_elem(unused_labels)
    v = _pop_random_elem(unused_labels)

    used_labels = [u, v]
    splits = [Split(part1={u}, part2={v})]
    while unused_labels:
        u = _pop_random_elem(unused_labels)

        divided_split = _pop_random_elem(splits)

        updated_splits = [
            Split(part1={u}, part2=used_labels),
            Split(part1=divided_split.part1 | {u}, part2=divided_split.part2),
            Split(part1=divided_split.part1, part2=divided_split.part2 | {u}),
        ]

        for split in splits:
            updated_splits.append(
                Split(
                    part1=split.get_part_away_from(divided_split),
                    part2=split.get_part_towards(divided_split) | {u},
                )
            )

        used_labels.append(u)
        splits = updated_splits
    return splits


def check_if_separated(labels, splits):
    """Check for each pair of labels if exists split that separates them.

    Parameters
    ----------
    labels : list[int]
        A list of integers, the set of labels that we generate splits for.
    splits : list[Split]
        A list of splits of the set of labels.

    Returns
    -------
    are_separated : bool
        True if the labels are pair-wise separated by a split else False.
    """
    return gs.all(
        [
            gs.any([sp.separates(u, v) for sp in splits])
            for u, v in itertools.combinations(labels, 2)
        ]
    )


def delete_splits(splits, labels, p_keep, check=True):
    """Delete splits randomly from a set of splits.

    We require the splits to satisfy the check for if all pair-wise labels are
    separated. In this way, before deleting a split, this condition is checked
    to make sure it is not violated.

    Parameters
    ----------
    splits : list[Split]
        A list of splits of the set of labels.
    labels : list[int]
        A list of integers, the set of labels that we generate splits for.
    p_keep : float
        A float between 0 and 1 determining the probability with which a split
        is kept and not deleted.
    check : bool
        If True, checks if splits still separate all labels. In this case, the split
        will not be deleted. If False, any split can be randomly deleted.

    Returns
    -------
    left_over_splits : list[Split]
        The list of splits that are not deleted.
    """
    if p_keep == 1:
        return splits

    for i in reversed(range(len(splits))):
        if gs.random.rand(1) > p_keep:
            splits_cp = splits.copy()
            splits_cp.pop(i)
            if not check:
                splits = splits_cp
            elif check_if_separated(splits=splits_cp, labels=labels):
                splits = splits_cp
    return splits


@functools.total_ordering
class Split:
    r"""Two-set partitions of sets.

    Two-set partitions of a smaller subset of :math:`\{0,...,n-1\}` are also allowed,
    where :math:`n` is a positive integer, which is not passed as an argument as it is
    nowhere needed.

    The parameters ``part1`` and ``part2`` are assigned to the attributes ``self.part1``
    and ``self.part2`` in a unique sorted way: the one that contains the smallest
    minimal value is assigned to ``self.part1`` for consistency.

    Parameters
    ----------
    part1 : iterable
        The first part of the split, an iterable that is a subset of
        :math:`\{0,\dots,n-1\}`. It may be empty, but must have empty intersection with
        ``part2``.
    part2 : iterable
        The second part of the split, an iterable that is a subset of
        :math:`\{0,\dots,n-1\}`. It may be empty, but must have empty intersection with
        ``part1``.
    """

    def __init__(self, part1, part2):
        part1, part2 = set(part1), set(part2)
        if part1 & part2:
            raise ValueError(
                f"A split consists of disjoint sets, those are not: {part1}, {part2}."
            )
        if part1 and part2:
            self.part1, self.part2 = (
                (part1, part2) if min(part1) < min(part2) else (part2, part1)
            )
        else:
            self.part1 = part1 or part2
            self.part2 = set()

    def __bool__(self):
        """Return True if and only if both parts are non-empty.

        We use the boolean representation to indicate whether both parts of a split are
        non-empty.

        Returns
        -------
        boolean_of_split : bool
            Returns the boolean representation of a split.
        """
        return bool(self.part1) and bool(self.part2)

    def __eq__(self, other):
        """Check for equal hashes of the two splits.

        Parameters
        ----------
        other : Split
            The other split.

        Returns
        -------
        is_equal : bool
            Return ``True`` if the splits are equal, else ``False``.
        """
        return hash(self) == hash(other)

    def __hash__(self):
        """Compute the hash of a split.

        Note that this hash simply uses the hash function for tuples.

        Returns
        -------
        hash_of_split : int
            Return the hash of the split.
        """
        return hash((tuple(self.part1), tuple(self.part2)))

    def __lt__(self, other):
        """Check if the hash of this split is less than the hash of the other split.

        Note that this partial ordering does not have a mathematical background, this is
        introduced in order to have a unique ordering for each set of splits at hand.

        Parameters
        ----------
        other : Split
            The other split.

        Returns
        -------
        is_strictly_less_than : bool
            Return ``True`` if hash is less than hash of other, else ``False``.
        """
        return hash(self) < hash(other)

    def __repr__(self):
        """Return the string representation of the split.

        This string representation requires that one can retrieve all necessary
        information from the string.

        Returns
        -------
        string_of_split : str
            Return the string representation of the split.
        """
        return str((self.part1, self.part2))

    def __str__(self):
        """Return the fancy printable string representation of the split.

        This string representation does NOT require that one can retrieve all necessary
        information from the string, but this string representation is required to be
        readable by a human.

        Returns
        -------
        string_of_split : str
            Return the fancy readable string representation of the split.
        """
        return f"{self.part1}|{self.part2}"

    def is_compatible(self, other):
        """Check whether this split is compatible with another split.

        Two splits are compatible, if at least one intersection of the respective parts
        of the splits is empty.

        Parameters
        ----------
        other : Split
            The other split.

        Returns
        -------
        is_compatible_with : bool
            Return ``True`` if the splits are compatible, else ``False``.
        """
        p1, p2 = self.part1, self.part2
        o1, o2 = other.part1, other.part2
        return sum([bool(s) for s in [p1 & o1, p1 & o2, p2 & o1, p2 & o2]]) < 4

    def get_part_away_from(self, other):
        """Return the part of this split that is directed away from other split.

        Parameters
        ----------
        other : Split
            The other split.

        Returns
        -------
        part_that_does_not_point : iterable
            Return the part of the split ``self`` that does not point toward
            ``other``. See ``self.get_part_towards`` for further explanation.
        """
        if other.part_contains(self.part1):
            return self.part1
        return self.part2

    def get_part_towards(self, other):
        """Return the part of this split that is directed toward the other split.

        Each split contains part1 and part2, the parts that the corresponding edge in
        the graph splits the set of labels into. Thus, one can think of the split as an
        edge, where part1 points in the direction of the part of the tree where the
        labels of part1 are contained, and part2 points in the other direction.
        So, part1 points in the direction of ``other``, if it corresponds to an
        edge that is contained in the tree that part1 points to, else part2 points in
        the direction of ``split``.

        Parameters
        ----------
        other : Split
            The other split.

        Returns
        -------
        part_towards : iterable
            Return the part of the split ``self`` that points toward ``other_split``.
        """
        if other.part_contains(self.part1):
            return self.part2
        return self.part1

    def part_contains(self, subset):
        """Determine if a subset is contained in either part of a split.

        Parameters
        ----------
        subset : set
            The subset containing labels.

        Returns
        -------
        is_contained : bool
            A boolean that is true if the subset is contained in ``self.part1`` or
            ``self.part2``.
        """
        return subset.issubset(self.part1) or subset.issubset(self.part2)

    def restrict_to(self, subset):
        r"""Return the restriction of a split to a subset.

        Parameters
        ----------
        subset : set
            The subset that the split is restricted to.

        Returns
        -------
        restr_split : Split
            The restricted split, if the split is :math:`A\vert B`, then the split
            restricted to the subset :math:`C` is :math:`A\cap C\vert B\cap C`.
        """
        return Split(
            part1=self.part1 & subset,
            part2=self.part2 & subset,
        )

    def separates(self, u, v):
        """Determine whether the labels (or label sets) are separated by the split.

        Parameters
        ----------
        u : list of int, int
            Either an integer or a set of labels.
        v : list of int, int
            Either an integer or a set of labels.

        Returns
        -------
        is_separated : bool
            A boolean determining whether u and v are separated by the split (i.e. if
             they are not in the same part).
        """
        u = {u} if isinstance(u, int) else set(u)
        v = {v} if isinstance(v, int) else set(v)
        b1 = u.issubset(self.part1) and v.issubset(self.part2)
        b2 = v.issubset(self.part1) and u.issubset(self.part2)
        return b1 or b2


class ForestTopology:
    r"""The topology of a forest, using a split-based graph-structure representation.

    A forest topology is a partition into non-empty sets of the set
    :math:`\{0,\dots,n-1\}`, together with a set of splits for each element of the
    partition, where every split is a two-set partition of the respective element.
    A structure basically describes a phylogenetic forest, where each set of splits
    gives the structure of the tree with the labels of the corresponding element of
    the partition.

    Parameters
    ----------
    partition : tuple
        A tuple of tuples that is a partition of the set :math:`\{0,\dots,n-1\}`,
        representing the label sets of each connected component of the forest topology.
    split_sets : tuple
        A tuple of tuples containing splits, where each set of splits contains only
        splits of the respective label set in the partition, so their order
        is related. The splits are the edges of the connected components of the forest,
        respectively, and thus the union of all sets of splits yields all edges of the
        forest topology.

    Attributes
    ----------
    n_labels : int
        Number of labels, the set of labels is then :math:`\{0,\dots,n-1\}`.
    n_splits : int
        Number of splits.
    where : dict
        Give the index of a split in the flattened list of all splits.
    sep : list of int
        An increasing list of numbers between 0 and m, where m is the total number
        of splits in ``self.split_sets``, starting with 0, where each number
        indicates that a new connected component starts at that index.
        Useful for example for unraveling the tuple of all splits into
        ``self.split_sets``.
    paths : list of dict
        A list of dictionaries, each dictionary is for the respective connected
        component of the forest, and the items of each dictionary are for each pair
        of labels u, v, u < v in the respective component, a list of the splits on the
        unique path between the labels u and v.
    support : list of array-like
        For each split, give an :math:`n\times n` dimensional matrix, where the
        uv-th entry is ``True`` if the split separates the labels u and v, else
        ``False``.
    """

    def __init__(self, partition, split_sets):
        self._check_init(partition, split_sets)

        self.n_labels = len(set.union(*[set(part) for part in partition]))
        partition = [tuple(sorted(x)) for x in partition]
        seq = [part[0] for part in partition]
        sort_key = sorted(range(len(seq)), key=seq.__getitem__)
        self.partition = tuple([partition[key] for key in sort_key])
        self.split_sets = tuple([tuple(sorted(split_sets[key])) for key in sort_key])

        self.where = {s: i for i, s in enumerate(self._flatten(self.split_sets))}

        lengths = [len(splits) for splits in self.split_sets]
        self.sep = [0] + [sum(lengths[0:j]) for j in range(1, len(lengths) + 1)]

        self.paths = [
            {
                (u, v): [s for s in splits if s.separates(u, v)]
                for u, v in itertools.combinations(part, r=2)
            }
            for part, splits in zip(self.partition, self.split_sets)
        ]

        _support = [
            gs.zeros((self.n_labels, self.n_labels), dtype=int)
            for _ in self._flatten(self.split_sets)
        ]
        for path_dict in self.paths:
            for (u, v), path in path_dict.items():
                for split in path:
                    _support[self.where[split]][u][v] = True
                    _support[self.where[split]][v][u] = True
        self.support = gs.reshape(
            gs.array([m for m in self._flatten(_support)]),
            (-1, self.n_labels, self.n_labels),
        )
        self._chart_gradient = None

        self.n_splits = gs.sum(
            gs.array([len(splits) for splits in self.split_sets]), dtype=int
        )

    def _check_init(self, partition, split_sets):
        if len(split_sets) != len(partition):
            raise ValueError(
                "Number of split sets is not equal to number of " "components."
            )
        for _part, _splits in zip(partition, split_sets):
            for _sp in _splits:
                if (_sp.part1 | _sp.part2) != set(_part):
                    raise ValueError(
                        f"The split {_sp} is not a split of component {_part}."
                    )

    def __eq__(self, other):
        """Check if ``self`` is equal to ``other``.

        Parameters
        ----------
        other : ForestTopology
            The other topology.

        Returns
        -------
        is_equal : bool
            Return ``True`` if the topologies are equal, else ``False``.
        """
        equal_n = self.n_labels == other.n_labels
        equal_partition = self.partition == other.partition
        equal_split_sets = self.split_sets == other.split_sets
        return equal_n and equal_partition and equal_split_sets

    def __ge__(self, other):
        """Check if ``self`` is greater than or equal to ``other``.

        Parameters
        ----------
        other : ForestTopology
            The other topology.

        Returns
        -------
        is_greater_than_or_equal : bool
            Return ``True`` if ``self`` is greater or equal to ``other``, else
            ``False``.
        """
        return other <= self

    def __gt__(self, other):
        """Check if ``self`` is strictly greater than ``other``.

        Parameters
        ----------
        other : ForestTopology
            The other topology.

        Returns
        -------
        is_greater_than : bool
            Return ``True`` if this topology is greater than the other, else ``False``.
        """
        return other < self

    def __hash__(self):
        """Compute the hash of a topology.

        Note that this hash simply uses the hash function for tuples.

        Returns
        -------
        hash_of_topology : int
            Return the hash of the topology.
        """
        return hash((self.n_labels, self.partition, self.split_sets))

    def __le__(self, other):
        """Check if ``self`` is less than or equal to ``other``.

        This partial ordering is the one defined in [1] and to show if self <= other is
        True, three things must be satisfied.
        (i)     ``self.partition`` must be a refinement of ``other.partition`` in the
                sense of partitions.
        (ii)    The splits of each component in ``self`` must be contained in the
                set of splits of ``other`` restricted to the component of ``self``.
        (iii)   Whenever two components of ``self`` are contained in a component of
                ``other``, there needs to exist a split in ``other`` separating those
                two components.
        If one of those three conditions are not fulfilled, this method returns False.

        Parameters
        ----------
        other : ForestTopology
            The structure to which self is compared to.

        Returns
        -------
        is_less_than_or_equal : bool
            Return ``True`` if (i), (ii) and (iii) are satisfied, else ``False``.
        """
        x_parts = [set(x) for x in self.partition]
        y_parts = [set(y) for y in other.partition]
        # (i)
        try:
            cover = {
                i: [j for j, y in enumerate(y_parts) if x.issubset(y)][0]
                for i, x in enumerate(x_parts)
            }
        except IndexError:
            return False
        # (ii)
        try:
            for (i, j), x in zip(cover.items(), x_parts):
                y_splits_restricted = {
                    split_y.restrict_to(subset=x) for split_y in other.split_sets[j]
                }
                if not set(self.split_sets[i]).issubset(y_splits_restricted):
                    raise NotPartialOrder()
        except NotPartialOrder:
            return False
        # (iii)
        try:
            for j in range(len(y_parts)):
                xs_in_y = [x for i, x in enumerate(x_parts) if cover[i] == j]
                for x1, x2 in itertools.combinations(xs_in_y, r=2):
                    sep_sp = [sp for sp in other.split_sets[j] if sp.separates(x1, x2)]
                    if not sep_sp:
                        raise NotPartialOrder()
        except NotPartialOrder:
            return False
        return True

    def __lt__(self, other):
        """Check if ``self`` is less than ``other``.

        Parameters
        ----------
        other : ForestTopology
            The other topology.

        Returns
        -------
        is_less_than : bool
            Return ``True`` if ``self`` less than ``other``, else ``False``.
        """
        return self <= other and self != other

    def __repr__(self):
        """Return the string representation of the topology.

        This string representation requires that one can retrieve all necessary
        information from the string.

        Returns
        -------
        string_of_topology : str
            Return the string representation of the topology.
        """
        return str((self.n_labels, self.partition, self.split_sets))

    def __str__(self):
        """Return the fancy printable string representation of the topology.

        This string representation does NOT require that one can retrieve all necessary
        information from the string, but this string representation is required to be
        readable by a human.

        Returns
        -------
        string_of_topology : str
            Return the fancy readable string representation of the topology.
        """
        comps = [", ".join(str(sp) for sp in splits) for splits in self.split_sets]
        return "(" + "; ".join(comps) + ")"

    def corr(self, weights):
        """Compute the correlation matrix of the topology with edge weights ``weights``.

        Parameters
        ----------
        weights : array-like, [n_splits]
            Edge weights.

        Returns
        -------
        corr : array-like, shape=[n, n]
            Returns the corresponding correlation matrix.
        """
        corr = gs.zeros((self.n_labels, self.n_labels))
        for path_dict in self.paths:
            for (u, v), path in path_dict.items():
                corr[u][v] = gs.prod(
                    gs.array([1 - weights[self.where[split]] for split in path])
                )
                corr[v][u] = corr[u][v]

        corr = gs.array(corr)
        return corr + gs.eye(corr.shape[0])

    def corr_gradient(self, weights):
        """Compute the gradient of the correlation matrix, differentiated by weights.

        Parameters
        ----------
        weights : array-like, [n_splits]
            The vector weights at which the gradient is computed.

        Returns
        -------
        gradient : array-like, shape=[n_splits, n, n]
            The gradient of the correlation matrix, differentiated by weights.
        """
        x_list = [
            [y if i != k else 0 for i, y in enumerate(weights)]
            for k in range(len(weights))
        ]
        gradient = gs.array(
            [-supp * self.corr(x) for supp, x in zip(self.support, x_list)]
        )
        return gradient

    def _unflatten(self, ls):
        """Transform list into list of lists according to separators, ``self.sep``.

        The separators are a list of integers, increasing. Then, all elements between to
        indices in separators will be put into a list, and together, all lists give a
        nested list.

        Parameters
        ----------
        ls : iterable
            The flat list that will be nested.

        Returns
        -------
        ls_nested : list[list]
            The nested list of lists.
        """
        return [ls[i:j] for i, j in zip(self.sep[:-1], self.sep[1:])]

    @staticmethod
    def _flatten(ls):
        """Flatten a list of lists into a single list by concatenation.

        Parameters
        ----------
        ls : nested list
            The nested list to flatten.

        Returns
        -------
        ls_flat : list, tuple
            The flatted list.
        """
        return [y for z in ls for y in z]



from geomstats.test.test_case import TestCase


class OptimizerTestCase(TestCase):
    def test_minimize(
        self,
        fun,
        x0,
        expected,
        atol,
        fun_jac=None,
        fun_hess=None,
        hessp=None,
    ):
        res = self.optimizer.minimize(
            fun, x0, fun_jac=fun_jac, fun_hess=fun_hess, hessp=hessp
        )
        self.assertAllClose(res.x, expected, atol=atol)


class RootFinderTestCase(TestCase):
    def test_root(
        self,
        fun,
        x0,
        expected,
        atol,
        fun_jac=None,
    ):
        res = self.root_finder.root(fun, x0, fun_jac=fun_jac)
        self.assertAllClose(res.x, expected, atol=atol)



"""Optimizers implementations."""

import logging
from abc import ABC, abstractmethod

import scipy
from scipy.optimize import OptimizeResult

import geomstats.backend as gs
from geomstats.exceptions import AutodiffNotImplementedError
from geomstats.numerics._common import result_to_backend_type


class ScipyMinimize:
    """Wrapper for scipy.optimize.minimize.

    Only `autodiff_jac` and `autodiff_hess` differ from scipy: if True, then
    automatic differentiation is used to compute jacobian and/or hessian.
    """

    def __init__(
        self,
        method="L-BFGS-B",
        autodiff_jac=False,
        autodiff_hess=False,
        bounds=None,
        constraints=(),
        tol=None,
        callback=None,
        options=None,
        save_result=False,
    ):
        if (autodiff_jac or autodiff_hess) and not gs.has_autodiff():
            raise AutodiffNotImplementedError(
                "Minimization with 'autodiff' requires automatic differentiation."
                "Change backend via the command "
                "export GEOMSTATS_BACKEND=pytorch in a terminal"
            )

        self.method = method
        self.autodiff_jac = autodiff_jac
        self.autodiff_hess = autodiff_hess
        self.bounds = bounds
        self.constraints = constraints
        self.tol = tol
        self.callback = callback
        self.options = options

        self.save_result = save_result
        self.result_ = None

    def _handle_jac(self, fun, fun_jac):
        if fun_jac is not None:
            fun_ = lambda x: fun(gs.from_numpy(x))
            fun_jac_ = fun_jac
            if callable(fun_jac):
                fun_jac_ = lambda x: fun_jac(gs.from_numpy(x))

            return fun_, fun_jac_

        if self.autodiff_jac:
            jac = True
            fun_ = lambda x: gs.autodiff.value_and_grad(fun)(gs.from_numpy(x))
        else:
            jac = fun_jac
            fun_ = lambda x: fun(gs.from_numpy(x))

        return fun_, jac

    def _handle_hess(self, fun, fun_hess):
        if fun_hess is not None or not self.autodiff_hess:
            fun_hess_ = fun_hess
            if callable(fun_hess):
                fun_hess_ = lambda x: fun_hess(gs.from_numpy(x))

            return fun_hess_

        return lambda x: gs.autodiff.hessian(fun)(gs.from_numpy(x))

    def minimize(self, fun, x0, fun_jac=None, fun_hess=None, hessp=None):
        """Minimize objective function.

        Parameters
        ----------
        fun : callable
            The objective function to be minimized.
        x0 : array-like
            Initial guess.
        fun_jac : callable
            Jacobian of fun.
        fun_hess : callable
            Hessian of fun.
        hessp : callable
        """
        fun_, jac = self._handle_jac(fun, fun_jac)
        hess = self._handle_hess(fun, fun_hess)

        result = scipy.optimize.minimize(
            fun_,
            x0,
            method=self.method,
            jac=jac,
            hess=hess,
            hessp=hessp,
            bounds=self.bounds,
            tol=self.tol,
            constraints=self.constraints,
            callback=self.callback,
            options=self.options,
        )

        result = result_to_backend_type(result)

        if not result.success:
            logging.warning(result.message)

        if self.save_result:
            self.result_ = result

        return result


class RootFinder(ABC):
    """Find a root of a vector-valued function."""

    @abstractmethod
    def root(self, fun, x0, fun_jac=None):
        """Find a root of a vector-valued function.

        Parameters
        ----------
        fun : callable
            Vector-valued function.
        x0 : array-like
            Initial guess.
        fun_jac : callable
            Ignored if None.

        Returns
        -------
        res : OptimizeResult
        """


class ScipyRoot(RootFinder):
    """Wrapper for scipy.optimize.root.

    Only `autodiff_jac` differs from scipy: if True, then
    automatic differentiation is used to compute jacobian.
    """

    def __init__(
        self,
        method="hybr",
        autodiff_jac=False,
        tol=None,
        callback=None,
        options=None,
        save_result=False,
    ):
        if (autodiff_jac) and not gs.has_autodiff():
            raise AutodiffNotImplementedError(
                "Root finding with 'autodiff' requires automatic differentiation."
                "Change backend via the command "
                "export GEOMSTATS_BACKEND=pytorch in a terminal"
            )

        self.method = method
        self.autodiff_jac = autodiff_jac
        self.tol = tol
        self.callback = callback
        self.options = options
        self.save_result = save_result
        self.result_ = None

    def _handle_jac(self, fun, fun_jac):
        if fun_jac is not None:
            fun_ = lambda x: fun(gs.from_numpy(x))
            fun_jac_ = fun_jac
            if callable(fun_jac):
                fun_jac_ = lambda x: fun_jac(gs.from_numpy(x))

            return fun_, fun_jac_

        if self.autodiff_jac:
            jac = True
            fun_ = lambda x: gs.autodiff.value_and_jacobian(fun)(gs.from_numpy(x))
        else:
            jac = fun_jac
            fun_ = lambda x: fun(gs.from_numpy(x))

        return fun_, jac

    def root(self, fun, x0, fun_jac=None):
        """Find a root of a vector-valued function.

        Parameters
        ----------
        fun : callable
            Vector-valued function.
        x0 : array-like
            Initial guess.
        fun_jac : callable
            Jacobian of fun. Ignored if None.

        Returns
        -------
        res : OptimizeResult
        """
        fun_, fun_jac_ = self._handle_jac(fun, fun_jac)

        result = scipy.optimize.root(
            fun_,
            x0,
            method=self.method,
            jac=fun_jac_,
            tol=self.tol,
            callback=self.callback,
            options=self.options,
        )

        result = result_to_backend_type(result)

        if not result.success:
            logging.warning(result.message)

        if self.save_result:
            self.result_ = result

        return result


class NewtonMethod(RootFinder):
    """Find a root of a vector-valued function with Newton's method.

    Check https://en.wikipedia.org/wiki/Newton%27s_method_in_optimization
    for details.

    Parameters
    ----------
    atol : float
        Tolerance to check algorithm convergence.
    max_iter : int
        Maximum iterations.
    """

    def __init__(self, atol=gs.atol, max_iter=100):
        self.atol = atol
        self.max_iter = max_iter

    def root(self, fun, x0, fun_jac):
        """Find a root of a vector-valued function.

        Parameters
        ----------
        fun : callable
            Vector-valued function.
        x0 : array-like
            Initial guess.
        fun_jac : callable
            Jacobian of fun.
        """
        xk = x0
        message = "The solution converged."
        status = 1
        for it in range(self.max_iter):
            fun_xk = fun(xk)
            if gs.linalg.norm(fun_xk) <= self.atol:
                break

            y = gs.linalg.solve(fun_jac(xk), fun_xk)
            xk = xk - y
        else:
            message = f"Maximum number of iterations {self.max_iter} reached. Results may be inaccurate"
            status = 0

        result = OptimizeResult(
            x=xk,
            success=(status == 1),
            status=status,
            method="newton",
            message=message,
            nfev=it + 1,
            njac=it,
        )

        if not result.success:
            logging.warning(result.message)

        return result



"""Frechet mean.

Lead authors: Nicolas Guigui and Nina Miolane.
"""

import abc
import logging
import math

from sklearn.base import BaseEstimator

import geomstats.backend as gs
import geomstats.errors as error
from geomstats.geometry.discrete_curves import ElasticMetric, SRVMetric
from geomstats.geometry.euclidean import EuclideanMetric
from geomstats.geometry.hypersphere import HypersphereMetric

ELASTIC_METRICS = [SRVMetric, ElasticMetric]


def _is_linear_metric(metric):
    return isinstance(metric, EuclideanMetric)


def _is_elastic_metric(metric):
    return isinstance(metric, tuple(ELASTIC_METRICS))


def _scalarmul(scalar, array):
    return gs.einsum("n,n...->n...", scalar, array)


def _scalarmulsum(scalar, array):
    return gs.einsum("n,n...->...", scalar, array)


def _batchscalarmulsum(array_1, array_2):
    return gs.einsum("ni,ni...->i...", array_1, array_2)


def variance(space, points, base_point, weights=None):
    """Variance of (weighted) points wrt a base point.

    Parameters
    ----------
    space : Manifold
        Equipped manifold.
    points : array-like, shape=[n_samples, dim]
        Points.
    weights : array-like, shape=[n_samples,]
        Weights associated to the points.
        Optional, default: None.

    Returns
    -------
    var : float
       Weighted variance of the points.
    """
    if weights is None:
        n_points = gs.shape(points)[0]
        weights = gs.ones((n_points,))

    sum_weights = gs.sum(weights)
    sq_dists = space.metric.squared_dist(base_point, points)
    var = weights * sq_dists

    var = gs.sum(var)
    var /= sum_weights

    return var


def linear_mean(points, weights=None):
    """Compute the weighted linear mean.

    The linear mean is the Frechet mean when points:

    - lie in a Euclidean space with Euclidean metric,
    - lie in a Minkowski space with Minkowski metric.

    Parameters
    ----------
    points : array-like, shape=[n_samples, dim]
        Points to be averaged.
    weights : array-like, shape=[n_samples,]
        Weights associated to the points.
        Optional, default: None.

    Returns
    -------
    mean : array-like, shape=[dim,]
        Weighted linear mean of the points.
    """
    if weights is None:
        n_points = gs.shape(points)[0]
        weights = gs.ones(n_points)
    sum_weights = gs.sum(weights)

    weighted_points = _scalarmul(weights, points)

    mean = gs.sum(weighted_points, axis=0) / sum_weights
    return mean


class BaseGradientDescent(abc.ABC):
    """Base class for gradient descent.

    Parameters
    ----------
    max_iter : int, optional
        Maximum number of iterations for the gradient descent.
    epsilon : float, optional
        Tolerance for stopping the gradient descent.
    init_point : array-like, shape=[*metric.shape]
        Initial point.
        Optional, default : None. In this case the first sample of the input
        data is used.
    init_step_size : float
        Learning rate in the gradient descent.
        Optional, default: 1.
    verbose : bool
        Level of verbosity to inform about convergence.
        Optional, default: False.
    """

    def __init__(
        self,
        max_iter=32,
        epsilon=1e-4,
        init_point=None,
        init_step_size=1.0,
        verbose=False,
    ):
        self.max_iter = max_iter
        self.epsilon = epsilon
        self.init_step_size = init_step_size
        self.verbose = verbose
        self.init_point = init_point

    @abc.abstractmethod
    def minimize(self, space, points, weights=None):
        """Perform gradient descent."""
        pass


class GradientDescent(BaseGradientDescent):
    """Default gradient descent."""

    def minimize(self, space, points, weights=None):
        """Perform default gradient descent."""
        n_points = gs.shape(points)[0]
        if weights is None:
            weights = gs.ones((n_points,))

        mean = points[0] if self.init_point is None else self.init_point

        if n_points == 1:
            return mean

        sum_weights = gs.sum(weights)
        iteration = 0
        sq_dist = 0.0
        var = 0.0

        norm_old = gs.linalg.norm(points)
        step_size = self.init_step_size

        while iteration < self.max_iter:
            logs = space.metric.log(point=points, base_point=mean)

            var = gs.sum(space.metric.squared_norm(logs, mean) * weights) / sum_weights

            tangent_mean = _scalarmulsum(weights, logs)
            tangent_mean /= sum_weights
            norm = gs.linalg.norm(tangent_mean)

            sq_dist = space.metric.squared_norm(tangent_mean, mean)

            var_is_0 = gs.isclose(var, 0.0)

            sq_dist_is_small = gs.less_equal(sq_dist, self.epsilon * space.dim)

            condition = ~gs.logical_or(var_is_0, sq_dist_is_small)
            if not (condition or iteration == 0):
                break

            estimate_next = space.metric.exp(step_size * tangent_mean, mean)
            mean = estimate_next
            iteration += 1

            if norm < norm_old:
                norm_old = norm
            elif norm > norm_old:
                step_size = step_size / 2.0

        if iteration == self.max_iter:
            logging.warning(
                "Maximum number of iterations %d reached. The mean may be inaccurate",
                self.max_iter,
            )

        if self.verbose:
            logging.info(
                "n_iter: {}, final variance: {}, final dist: {}".format(
                    iteration, var, sq_dist
                )
            )

        return mean


class BatchGradientDescent(BaseGradientDescent):
    """Batch gradient descent."""

    def minimize(self, space, points, weights=None):
        """Perform batch gradient descent."""
        shape = points.shape
        n_points, n_batch = shape[:2]
        point_shape = shape[2:]

        if n_points == 1:
            return points[0]

        if weights is None:
            weights = gs.ones((n_points, n_batch))

        flat_shape = (n_batch * n_points,) + point_shape
        estimates = points[0] if self.init_point is None else self.init_point
        points_flattened = gs.reshape(points, (n_points * n_batch,) + point_shape)
        convergence = math.inf
        iteration = 0
        convergence_old = convergence

        step_size = self.init_step_size

        while convergence > self.epsilon and self.max_iter > iteration:
            iteration += 1
            estimates_broadcast, _ = gs.broadcast_arrays(estimates, points)
            estimates_flattened = gs.reshape(estimates_broadcast, flat_shape)

            tangent_grad = space.metric.log(points_flattened, estimates_flattened)
            tangent_grad = gs.reshape(tangent_grad, shape)

            tangent_mean = _batchscalarmulsum(weights, tangent_grad) / n_points

            next_estimates = space.metric.exp(step_size * tangent_mean, estimates)
            convergence = gs.sum(space.metric.squared_norm(tangent_mean, estimates))
            estimates = next_estimates

            if convergence < convergence_old:
                convergence_old = convergence
            elif convergence > convergence_old:
                step_size = step_size / 2.0

        if iteration == self.max_iter:
            logging.warning(
                "Maximum number of iterations %d reached. The mean may be inaccurate",
                self.max_iter,
            )

        if self.verbose:
            logging.info(
                "n_iter: %d, final dist: %e, final step size: %e",
                iteration,
                convergence,
                step_size,
            )

        return estimates


class AdaptiveGradientDescent(BaseGradientDescent):
    """Adaptive gradient descent."""

    def minimize(self, space, points, weights=None):
        """Perform adaptive gradient descent.

        Frechet mean of (weighted) points using adaptive time-steps
        The loss function optimized is :math:`||M_1(x)||_x`
        (where :math:`M_1(x)` is the tangent mean at x) rather than
        the mean-square-distance (MSD) because this simplifies computations.
        Adaptivity is done in a Levenberg-Marquardt style weighting variable tau
        between the first order and the second order Gauss-Newton gradient descent.

        Parameters
        ----------
        points : array-like, shape=[n_samples, *metric.shape]
            Points to be averaged.
        weights : array-like, shape=[n_samples,], optional
            Weights associated to the points.

        Returns
        -------
        current_mean: array-like, shape=[*metric.shape]
            Weighted Frechet mean of the points.
        """
        n_points = gs.shape(points)[0]

        tau_max = 1e6
        tau_mul_up = 1.6511111
        tau_min = 1e-6
        tau_mul_down = 0.1

        if n_points == 1:
            return points[0]

        current_mean = points[0] if self.init_point is None else self.init_point

        if weights is None:
            weights = gs.ones((n_points,))
        sum_weights = gs.sum(weights)

        tau = self.init_step_size
        iteration = 0

        logs = space.metric.log(point=points, base_point=current_mean)
        var = (
            gs.sum(space.metric.squared_norm(logs, current_mean) * weights)
            / sum_weights
        )

        current_tangent_mean = _scalarmulsum(weights, logs)
        current_tangent_mean /= sum_weights
        sq_norm_current_tangent_mean = space.metric.squared_norm(
            current_tangent_mean, base_point=current_mean
        )

        while (
            sq_norm_current_tangent_mean > self.epsilon**2 and iteration < self.max_iter
        ):
            iteration += 1

            shooting_vector = tau * current_tangent_mean
            next_mean = space.metric.exp(
                tangent_vec=shooting_vector, base_point=current_mean
            )

            logs = space.metric.log(point=points, base_point=next_mean)
            var = (
                gs.sum(space.metric.squared_norm(logs, current_mean) * weights)
                / sum_weights
            )

            next_tangent_mean = _scalarmulsum(weights, logs)
            next_tangent_mean /= sum_weights
            sq_norm_next_tangent_mean = space.metric.squared_norm(
                next_tangent_mean, base_point=next_mean
            )

            if sq_norm_next_tangent_mean < sq_norm_current_tangent_mean:
                current_mean = next_mean
                current_tangent_mean = next_tangent_mean
                sq_norm_current_tangent_mean = sq_norm_next_tangent_mean
                tau = min(tau_max, tau_mul_up * tau)
            else:
                tau = max(tau_min, tau_mul_down * tau)

        if iteration == self.max_iter:
            logging.warning(
                "Maximum number of iterations %d reached. The mean may be inaccurate",
                self.max_iter,
            )

        if self.verbose:
            logging.info(
                "n_iter: %d, final variance: %e, final dist: %e, final_step_size: %e",
                iteration,
                var,
                sq_norm_current_tangent_mean,
                tau,
            )

        return current_mean


class LinearMean(BaseEstimator):
    """Linear mean.

    Parameters
    ----------
    space : Manifold
        Equipped manifold.

    Attributes
    ----------
    estimate_ : array-like, shape=[*space.shape]
        If fit, Frechet mean.
    """

    def __init__(self, space):
        self.space = space
        self.estimate_ = None

    def fit(self, X, y=None, weights=None):
        """Compute the Euclidean mean.

        Parameters
        ----------
        X : array-like, shape=[n_samples, *metric.shape]
            Training input samples.
        y : None
            Target values. Ignored.
        weights : array-like, shape=[n_samples,]
            Weights associated to the samples.
            Optional, default: None, in which case it is equally weighted.

        Returns
        -------
        self : object
            Returns self.
        """
        self.estimate_ = linear_mean(points=X, weights=weights)
        return self


class ElasticMean(BaseEstimator):
    """Elastic mean.

    Parameters
    ----------
    space : Manifold
        Equipped manifold.

    Attributes
    ----------
    estimate_ : array-like, shape=[*space.shape]
        If fit, Frechet mean.
    """

    def __init__(self, space):
        self.space = space
        self.estimate_ = None

    def _elastic_mean(self, points, weights=None):
        """Compute the weighted mean of elastic curves.

        SRV: Square Root Velocity.

        SRV curves are a special case of Elastic curves.

        The computation of the mean goes as follows:

        - Transform the curves into their SRVs/F-transform representations,
        - Compute the linear mean of the SRVs/F-transform representations,
        - Inverse-transform the mean in curve space.

        Parameters
        ----------
        points : array-like, shape=[n_samples, k_sampling_points, dim]
            Points on the manifold of curves (i.e. curves) to be averaged.
        weights : array-like, shape=[n_samples,]
            Weights associated to the points (i.e. curves).
            Optional, default: None.

        Returns
        -------
        mean : array-like, shape=[k_sampling_points, dim]
            Weighted linear mean of the points (i.e. of the curves).
        """
        diffeo = self.space.metric.diffeo
        transformed = diffeo(points)
        transformed_linear_mean = linear_mean(transformed, weights=weights)

        return diffeo.inverse(transformed_linear_mean)

    def fit(self, X, y=None, weights=None):
        """Compute the elastic mean.

        Parameters
        ----------
        X : array-like, shape=[n_samples, *metric.shape]
            Training input samples.
        y : None
            Target values. Ignored.
        weights : array-like, shape=[n_samples,]
            Weights associated to the samples.
            Optional, default: None, in which case it is equally weighted.

        Returns
        -------
        self : object
            Returns self.
        """
        self.estimate_ = self._elastic_mean(X, weights=weights)
        return self


class CircleMean(BaseEstimator):
    """Circle mean.

    Parameters
    ----------
    space : Manifold
        Equipped manifold.

    Attributes
    ----------
    estimate_ : array-like, shape=[2,]
        If fit, Frechet mean.
    """

    def __init__(self, space):
        self.space = space
        self.estimate_ = None

    def _circle_mean(self, points):
        """Determine the mean on a circle.

        Data are expected in radians in the range [-pi, pi). The mean is returned
        in the same range. If the mean is unique, this algorithm is guaranteed to
        find it. It is not vulnerable to local minima of the Frechet function. If
        the mean is not unique, the algorithm only returns one of the means. Which
        mean is returned depends on numerical rounding errors.

        Parameters
        ----------
        points : array-like, shape=[n_samples, 1]
            Data set of angles (intrinsic coordinates).

        Reference
        ---------
        .. [HH15] Hotz, T. and S. F. Huckemann (2015), "Intrinsic means on the
            circle: Uniqueness, locus and asymptotics", Annals of the Institute of
            Statistical Mathematics 67 (1), 177–193.
            https://arxiv.org/abs/1108.2141
        """
        sample_size = points.shape[0]
        mean0 = gs.mean(points)
        var0 = gs.sum((points - mean0) ** 2)
        sorted_points = gs.sort(points, axis=0)
        means = self._circle_variances(mean0, var0, sample_size, sorted_points)
        return means[gs.argmin(means[:, 1]), 0]

    @staticmethod
    def _circle_variances(mean, var, n_samples, points):
        """Compute the minimizer of the variance functional.

        Parameters
        ----------
        mean : float
            Mean angle.
        var : float
            Variance of the angles.
        n_samples : int
            Number of samples.
        points : array-like, shape=[n_samples,]
            Data set of ordered angles.

        References
        ----------
        .. [HH15] Hotz, T. and S. F. Huckemann (2015), "Intrinsic means on the
            circle: Uniqueness, locus and asymptotics", Annals of the Institute of
            Statistical Mathematics 67 (1), 177–193.
            https://arxiv.org/abs/1108.2141
        """
        means = (mean + gs.linspace(0.0, 2 * gs.pi, n_samples + 1)[:-1]) % (2 * gs.pi)
        means = gs.where(means >= gs.pi, means - 2 * gs.pi, means)
        parts = gs.array([gs.sum(points) / n_samples if means[0] < 0 else 0])
        m_plus = means >= 0
        left_sums = gs.cumsum(points)
        right_sums = left_sums[-1] - left_sums
        i = gs.arange(n_samples, dtype=right_sums.dtype)
        j = i[1:]
        parts2 = right_sums[:-1] / (n_samples - j)
        first_term = parts2[:1]
        parts2 = gs.where(m_plus[1:], left_sums[:-1] / j, parts2)
        parts = gs.concatenate([parts, first_term, parts2[1:]])

        # Formula (6) from [HH15]_
        plus_vec = (4 * gs.pi * i / n_samples) * (gs.pi + parts - mean) - (
            2 * gs.pi * i / n_samples
        ) ** 2
        minus_vec = (4 * gs.pi * (n_samples - i) / n_samples) * (
            gs.pi - parts + mean
        ) - (2 * gs.pi * (n_samples - i) / n_samples) ** 2
        minus_vec = gs.where(m_plus, plus_vec, minus_vec)
        means = gs.transpose(gs.vstack([means, var + minus_vec]))
        return means

    def fit(self, X, y=None):
        """Compute the circle mean.

        Parameters
        ----------
        X : array-like, shape=[n_samples, 2]
            Training input samples.
        y : None
            Target values. Ignored.
        weights : array-like, shape=[n_samples,]
            Weights associated to the samples.
            Optional, default: None, in which case it is equally weighted.

        Returns
        -------
        self : object
            Returns self.
        """
        self.estimate_ = self.space.angle_to_extrinsic(
            self._circle_mean(self.space.extrinsic_to_angle(X))
        )
        return self


class FrechetMean(BaseEstimator):
    r"""Empirical Frechet mean.

    Parameters
    ----------
    space : Manifold
        Equipped manifold.
    method : str, {\'default\', \'adaptive\', \'batch\'}
        Gradient descent method.
        The `adaptive` method uses a Levenberg-Marquardt style adaptation of
        the learning rate. The `batch` method is similar to the default
        method but for batches of equal length of samples. In this case,
        samples must be of shape [n_samples, n_batch, *space.shape].
        Optional, default: \'default\'.

    Attributes
    ----------
    estimate_ : array-like, shape=[*space.shape]
        If fit, Frechet mean.

    Notes
    -----
    * Required metric methods for general case:
        * `log`, `exp`, `squared_norm` (for convergence criteria)
    """

    def __new__(cls, space, **kwargs):
        """Interface for instantiating proper algorithm."""
        if isinstance(space.metric, HypersphereMetric) and space.dim == 1:
            return CircleMean(space, **kwargs)

        elif _is_linear_metric(space.metric):
            return LinearMean(space, **kwargs)

        elif _is_elastic_metric(space.metric):
            return ElasticMean(space, **kwargs)

        return super().__new__(cls)

    def __init__(self, space, method="default"):
        self.space = space

        self._method = None
        self.method = method

        self.estimate_ = None

    def set(self, **kwargs):
        """Set optimizer parameters.

        Especially useful for one line instantiations.
        """
        for param_name, value in kwargs.items():
            if not hasattr(self.optimizer, param_name):
                raise ValueError(f"Unknown parameter {param_name}.")

            setattr(self.optimizer, param_name, value)
        return self

    @property
    def method(self):
        """Gradient descent method."""
        return self._method

    @method.setter
    def method(self, value):
        """Gradient descent method."""
        error.check_parameter_accepted_values(
            value, "method", ["default", "adaptive", "batch"]
        )
        if value == self._method:
            return

        self._method = value
        MAP_OPTIMIZER = {
            "default": GradientDescent,
            "adaptive": AdaptiveGradientDescent,
            "batch": BatchGradientDescent,
        }
        self.optimizer = MAP_OPTIMIZER[value]()

    def fit(self, X, y=None, weights=None):
        """Compute the empirical weighted Frechet mean.

        Parameters
        ----------
        X : array-like, shape=[n_samples, *metric.shape]
            Training input samples.
        y : None
            Target values. Ignored.
        weights : array-like, shape=[n_samples,]
            Weights associated to the samples.
            Optional, default: None, in which case it is equally weighted.

        Returns
        -------
        self : object
            Returns self.
        """
        self.estimate_ = self.optimizer.minimize(
            space=self.space,
            points=X,
            weights=weights,
        )
        return self



"""The vector space of lower triangular matrices.

Lead author: Saiteja Utpala.
"""

import geomstats.backend as gs
from geomstats.geometry.base import LevelSet, MatrixVectorSpace
from geomstats.geometry.matrices import Matrices, MatricesMetric


class LowerTriangularMatrices(MatrixVectorSpace):
    """Class for the vector space of lower triangular matrices of size n.

    Parameters
    ----------
    n : int
        Integer representing the shapes of the matrices: n x n.
    """

    def __init__(self, n, equip=True):
        super().__init__(dim=int(n * (n + 1) / 2), shape=(n, n), equip=equip)
        self.n = n

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return MatricesMetric

    def _create_basis(self):
        """Compute the basis of the vector space of lower triangular.

        Returns
        -------
        basis : array-like, shape=[dim, n, n]
            Basis matrices of the space.
        """
        tril_idxs = gs.ravel_tril_indices(self.n)
        vector_bases = gs.cast(
            gs.one_hot(tril_idxs, self.n * self.n),
            dtype=gs.get_default_dtype(),
        )
        return gs.reshape(vector_bases, [-1, self.n, self.n])

    def belongs(self, point, atol=gs.atol):
        """Evaluate if a matrix is lower triangular.

        Parameters
        ----------
        point : array-like, shape=[.., n, n]
            Point to test.
        atol : float
            Tolerance to evaluate equality with the transpose.

        Returns
        -------
        belongs : array-like, shape=[...,]
            Boolean evaluating if point belongs to the space.
        """
        belongs = super().belongs(point)
        if gs.any(belongs):
            is_lower_triangular = Matrices.is_lower_triangular(point, atol)
            return gs.logical_and(belongs, is_lower_triangular)
        return belongs

    @staticmethod
    def basis_representation(matrix_representation):
        """Convert a lower triangular matrix into a vector.

        Parameters
        ----------
        matrix_representation : array-like, shape=[..., n, n]
            Matrix.

        Returns
        -------
        vec : array-like, shape=[..., n(n+1)/2]
            Vector.
        """
        return gs.tril_to_vec(matrix_representation)

    def projection(self, point):
        """Make a square matrix lower triangular by zeroing out other elements.

        Parameters
        ----------
        point : array-like, shape=[..., n, n]
            Matrix.

        Returns
        -------
        sym : array-like, shape=[..., n, n]
            Symmetric matrix.
        """
        return gs.tril(point)


class StrictlyLowerTriangularMatrices(LevelSet, MatrixVectorSpace):
    r"""Strictly lower triangular matrices.

    Set of lower triangular matrices with null diagonal:

    .. math::

        \operatorname{LT}^0(n)=\{L \in \operatorname{LT}(n)
        \mid \operatorname{Diag}(L)=0\}

    Parameters
    ----------
    n : int
        Integer representing the shapes of the matrices: n x n.
    equip : bool
        If True, equip space with default metric.
    """

    def __init__(self, n, equip=True):
        self.n = n
        super().__init__(dim=int(n * (n - 1) / 2), equip=equip)

    def _define_embedding_space(self):
        """Define embedding space of the manifold.

        Returns
        -------
        embedding_space : Manifold
            Instance of Manifold.
        """
        return LowerTriangularMatrices(n=self.n)

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return MatricesMetric

    def submersion(self, point):
        """Submersion that defines the manifold.

        Parameters
        ----------
        point : array-like, shape=[..., n, n]

        Returns
        -------
        submersed_point : array-like, shape=[..., n]
        """
        return Matrices.diagonal(point)

    def tangent_submersion(self, vector, point):
        """Tangent submersion.

        Parameters
        ----------
        vector : array-like, shape=[..., n, n]
        point : Ignored.

        Returns
        -------
        submersed_vector : array-like, shape=[..., n]
        """
        submersed_vector = Matrices.diagonal(vector)
        if point is not None and point.ndim > vector.ndim:
            return gs.broadcast_to(submersed_vector, point.shape[:-1])

        return submersed_vector

    def _create_basis(self):
        """Create a basis for the vector space.

        Returns
        -------
        basis : array-like, shape=[dim, n, n]
            Basis matrices of the space.
        """
        tril_idxs = gs.ravel_tril_indices(self.n, k=-1)
        vector_bases = gs.cast(
            gs.one_hot(tril_idxs, self.n * self.n),
            dtype=gs.get_default_dtype(),
        )
        return gs.reshape(vector_bases, [-1, self.n, self.n])

    @staticmethod
    def basis_representation(matrix_representation):
        """Compute the coefficients of matrices in the given basis.

        Parameters
        ----------
        matrix_representation : array-like, shape=[..., n, n]
            Matrix.

        Returns
        -------
        vec : array-like, shape=[..., dim]
            Vector.
        """
        return gs.tril_to_vec(matrix_representation, k=-1)

    def projection(self, point):
        """Project a point to the manifold.

        Parameters
        ----------
        point: array-like, shape[..., n, n]
            Point to project.

        Returns
        -------
        proj_point: array-like, shape[..., n, n]
            Projected point.
        """
        proj_point = self.embedding_space.projection(point)
        return proj_point - gs.vec_to_diag(gs.diagonal(proj_point, axis1=-2, axis2=-1))



"""Autograd based linear algebra backend."""

import functools as _functools

import autograd.numpy as _np
from autograd.extend import defvjp as _defvjp
from autograd.extend import primitive as _primitive
from autograd.numpy.linalg import (
    cholesky,
    det,
    eig,
    eigh,
    eigvalsh,
    inv,
    matrix_power,
    matrix_rank,
    norm,
    svd,
)
from autograd.scipy.linalg import expm
from scipy.optimize import quadratic_assignment as _quadratic_assignment

from .._shared_numpy.linalg import (
    fractional_matrix_power,
    is_single_matrix_pd,
    polar,
    qr,
    solve,
    solve_sylvester,
    sqrtm,
)
from .._shared_numpy.linalg import logm as _logm


def _adjoint(_ans, x, fn):
    vectorized = x.ndim == 3
    axes = (0, 2, 1) if vectorized else (1, 0)

    def vjp(g):
        n = x.shape[-1]
        size_m = x.shape[:-2] + (2 * n, 2 * n)
        mat = _np.zeros(size_m)
        mat[..., :n, :n] = x.transpose(axes)
        mat[..., n:, n:] = x.transpose(axes)
        mat[..., :n, n:] = g
        return fn(mat)[..., :n, n:]

    return vjp


_expm_vjp = _functools.partial(_adjoint, fn=expm)
_defvjp(expm, _expm_vjp)


logm = _primitive(_logm)

_logm_vjp = _functools.partial(_adjoint, fn=logm)
_defvjp(logm, _logm_vjp)


def quadratic_assignment(a, b, options):
    return list(_quadratic_assignment(a, b, options=options).col_ind)



"""Wrapper around autograd functions to be consistent with backends."""

import autograd as _autograd
import autograd.numpy as _np
from autograd import jacobian


def _get_max_ndim_point(*points):
    """Identify point with higher dimension.

    Same as `geomstats.vectorization._get_max_ndim_point`.

    Parameters
    ----------
    points : array-like

    Returns
    -------
    max_ndim_point : array-like
        Point with higher dimension.
    """
    max_ndim_point = points[0]
    for point in points[1:]:
        if point.ndim > max_ndim_point.ndim:
            max_ndim_point = point

    return max_ndim_point


def _get_batch_shape(*points, point_ndims=1):
    """Get batch shape.

    Similar to `geomstats.vectorization.get_batch_shape`.

    Parameters
    ----------
    points : array-like or None
        Point belonging to the space.
    point_ndims : int or tuple[int]
        Point number of array dimensions.

    Returns
    -------
    batch_shape : tuple
        Returns the shape related with batch. () if only one point.
    """
    if isinstance(point_ndims, int):
        point_max_ndim = _get_max_ndim_point(*points)
        return point_max_ndim.shape[:-point_ndims]

    for point, point_ndim in zip(points, point_ndims):
        if point.ndim > point_ndim:
            return point.shape[:-point_ndim]

    return ()


def custom_gradient(*grad_funcs):
    """Create a decorator that allows a function to define its custom gradient(s).

    Parameters
    ----------
    *grad_funcs : callables
        Custom gradient functions.

    Returns
    -------
    decorator : callable
        This decorator, used on any function func, associates the
        input grad_funcs as the gradients of func.
    """

    def decorator(func):
        """Decorate a function to define its custome gradient(s).

        Parameters
        ----------
        func : callable
            Function whose gradients will be assigned by grad_funcs.

        Returns
        -------
        wrapped_function : callable
            Function func with gradients specified by grad_funcs.
        """
        wrapped_function = _autograd.extend.primitive(func)

        def wrapped_grad_func(i, ans, *args, **kwargs):
            grads = grad_funcs[i](*args, **kwargs)
            if isinstance(grads, float):
                return lambda g: g * grads
            if grads.ndim == 2:
                return lambda g: g[..., None] * grads
            if grads.ndim == 3:
                return lambda g: g[..., None, None] * grads
            return lambda g: g * grads

        if len(grad_funcs) == 1:
            _autograd.extend.defvjp(
                wrapped_function,
                lambda ans, *args, **kwargs: wrapped_grad_func(0, ans, *args, **kwargs),
            )
        elif len(grad_funcs) == 2:
            _autograd.extend.defvjp(
                wrapped_function,
                lambda ans, *args, **kwargs: wrapped_grad_func(0, ans, *args, **kwargs),
                lambda ans, *args, **kwargs: wrapped_grad_func(1, ans, *args, **kwargs),
            )
        elif len(grad_funcs) == 3:
            _autograd.extend.defvjp(
                wrapped_function,
                lambda ans, *args, **kwargs: wrapped_grad_func(0, ans, *args, **kwargs),
                lambda ans, *args, **kwargs: wrapped_grad_func(1, ans, *args, **kwargs),
                lambda ans, *args, **kwargs: wrapped_grad_func(2, ans, *args, **kwargs),
            )
        else:
            raise NotImplementedError(
                "custom_gradient is not yet implemented " "for more than 3 gradients."
            )

        return wrapped_function

    return decorator


def _grad(func, argnums=0):
    def _wrapped_grad(*x, **kwargs):
        if not hasattr(x[0], "ndim") or x[0].ndim < 2:
            return _autograd.grad(func, argnum=argnums)(*x, **kwargs)

        return _autograd.elementwise_grad(func, argnum=argnums)(*x, **kwargs)

    return _wrapped_grad


@_autograd.differential_operators.unary_to_nary
def _elementwise_value_and_grad(fun, x):
    # same as autograd.elementwise_grad, but also returning ans
    vjp, ans = _autograd.differential_operators._make_vjp(fun, x)
    if _autograd.differential_operators.vspace(ans).iscomplex:
        raise TypeError("Elementwise_grad only applies to real-output functions.")

    return ans, vjp(_autograd.differential_operators.vspace(ans).ones())


def value_and_grad(func, argnums=0, point_ndims=1):
    """Wrap autograd value_and_grad function.

    Suitable for use in scipy.optimize.

    Parameters
    ----------
    func : callable
        Function whose value and gradient values
        will be computed.
    argnums: int or tuple[int]
        Specifies arguments to compute gradients with respect to.
    point_ndims: int or tuple[int]
        Specifies arguments ndim.

    Returns
    -------
    value_and_grad : callable
        Function that returns func's value and
        func's gradients' values at its inputs args.
    """

    def _value_and_grad(*inputs, **kwargs):
        batch_shape = _get_batch_shape(*inputs, point_ndims=point_ndims)
        if len(batch_shape) == 0:
            return _autograd.value_and_grad(func, argnum=argnums)(*inputs, **kwargs)

        if len(inputs) > 1:
            point_ndims_ = (
                (point_ndims,) * len(inputs)
                if isinstance(point_ndims, int)
                else point_ndims
            )
            inputs_ = []
            for point, point_ndim in zip(inputs, point_ndims_):
                if point.shape[:-point_ndim] != batch_shape:
                    point = _autograd.numpy.broadcast_to(
                        point, batch_shape + point.shape
                    )
                inputs_.append(point)
            inputs = inputs_

        return _elementwise_value_and_grad(func, argnum=argnums)(*inputs, **kwargs)

    return _value_and_grad


@_autograd.differential_operators.unary_to_nary
def _value_and_jacobian_op(fun, x):
    # same as autograd.jacobian, but also returning ans
    vjp, ans = _autograd.differential_operators._make_vjp(fun, x)
    ans_vspace = _autograd.differential_operators.vspace(ans)
    jacobian_shape = ans_vspace.shape + _autograd.differential_operators.vspace(x).shape
    grads = map(vjp, ans_vspace.standard_basis())
    return ans, _np.reshape(_np.stack(grads), jacobian_shape)


def value_and_jacobian(fun, point_ndim=1):
    def _value_and_jacobian_vec(x):
        if x.ndim == point_ndim:
            return _value_and_jacobian_op(fun)(x)

        ans = []
        jac = []
        for one_x in x:
            ans_, jac_ = _value_and_jacobian_op(fun)(one_x)
            ans.append(ans_)
            jac.append(jac_)

        return _np.stack(ans), _np.stack(jac)

    return _value_and_jacobian_vec


def jacobian_vec(fun, point_ndim=1):
    """Wrap autograd jacobian function.

    We note that the jacobian function of autograd is not vectorized
    by default, thus we modify its behavior here.

    Default autograd behavior:

    If the jacobian for one point of shape (2,) is of shape (3, 2),
    then calling the jacobian on 4 points with shape (4, 2) will
    be of shape (3, 2, 4, 2).

    Modified behavior:

    Calling the jacobian on 4 points gives a tensor of shape (4, 3, 2).

    We use a for-loop to allow this function to be vectorized with
    respect to several inputs in point, because the flag vectorize=True
    fails.

    Parameters
    ----------
    fun : callable
        Function whose jacobian values
        will be computed.

    Returns
    -------
    func_with_jacobian : callable
        Function that returns func's jacobian
        values at its inputs args.
    """

    def _jac(x):
        if x.ndim == point_ndim:
            return jacobian(fun)(x)
        return _np.stack([jacobian(fun)(one_x) for one_x in x])

    return _jac


def hessian(fun, func_out_ndim=None):
    """Wrap autograd hessian function.

    For consistency with the other backend, we convert this to a tensor
    of shape (dim, dim).

    Parameters
    ----------
    func : callable
        Function whose hessian values
        will be computed.
    func_out_ndim : int
        Unused. Here for API consistency.

    Returns
    -------
    func_with_hessian : callable
        Function that returns func's hessian
        values at its inputs args.
    """

    def _hess(x):
        return _autograd.hessian(fun)(x)

    return _hess


def hessian_vec(func, point_ndim=1, func_out_ndim=None):
    """Wrap autograd hessian function.

    We note that the hessian function of autograd is not vectorized
    by default, thus we modify its behavior here.

    We force the hessian to return a tensor of shape (n_points, dim, dim)
    when several points are given as inputs.

    Parameters
    ----------
    func : callable
        Function whose hessian values
        will be computed.
    func_out_ndim : int
        Unused. Here for API consistency.

    Returns
    -------
    func_with_hessian : callable
        Function that returns func's hessian
        values at its inputs args.
    """
    hessian_func = hessian(func)

    def _hess(x):
        if x.ndim == point_ndim:
            return hessian_func(x)
        return _np.stack([hessian_func(one_x) for one_x in x])

    return _hess


def jacobian_and_hessian(func, func_out_ndim=None):
    """Wrap autograd jacobian and hessian functions.

    Parameters
    ----------
    func : callable
        Function whose jacobian and hessian values
        will be computed.
    func_out_ndim : int
        Unused. Here for API consistency.

    Returns
    -------
    func_with_jacobian_and_hessian : callable
        Function that returns func's jacobian and
        func's hessian values at its inputs args.
    """
    return value_and_jacobian(jacobian_vec(func))


def value_jacobian_and_hessian(func, func_out_ndim=None):
    """Compute value, jacobian and hessian.

    func is only called once.

    Parameters
    ----------
    func : callable
        Function whose jacobian and hessian values
        will be computed.
    func_out_ndim : int
        Unused. Here for API consistency.
    """
    cache = []

    def _cached_value_and_jacobian(fun, return_cached=False):
        def _jac(x):
            ans, jac = value_and_jacobian(fun)(x)
            if not return_cached:
                cache.append(ans)
                return jac

            value = _np.stack(cache)._value if len(cache) > 1 else cache[0]._value
            cache.clear()

            return value, ans, jac

        return _jac

    return _cached_value_and_jacobian(
        _cached_value_and_jacobian(func), return_cached=True
    )



"""Numpy based random backend."""

import numpy as _np
from numpy.random import default_rng as _default_rng
from numpy.random import randint, seed

from .._shared_numpy.random import choice, multivariate_normal, normal, rand, uniform



import geomstats.backend as gs
from geomstats.geometry.base import ComplexMatrixVectorSpace
from geomstats.geometry.euclidean import Euclidean
from geomstats.geometry.general_linear import SquareMatrices
from geomstats.vectorization import get_n_points


def get_random_quaternion(n_points=1):
    # https://stackoverflow.com/a/44031492/11011913
    size = (3, n_points) if n_points > 1 else 3
    u, v, w = gs.random.uniform(size=size)

    return gs.transpose(
        gs.array(
            [
                gs.sqrt(1 - u) * gs.sin(2 * gs.pi * v),
                gs.sqrt(1 - u) * gs.cos(2 * gs.pi * v),
                gs.sqrt(u) * gs.sin(2 * gs.pi * w),
                gs.sqrt(u) * gs.cos(2 * gs.pi * w),
            ]
        )
    )


def get_random_times(n_times, sort=True):
    if n_times == 1:
        return gs.random.rand(1)[0]

    time = gs.random.rand(n_times)
    if sort:
        return gs.sort(time)

    return time


class RandomDataGenerator:
    def __init__(self, space, amplitude=2.0):
        self.space = space
        self.amplitude = amplitude

    def random_point(self, n_points=1):
        return self.space.random_point(n_points)

    def random_tangent_vec(self, base_point):
        tangent_vec = self.space.random_tangent_vec(base_point)
        return tangent_vec / self.amplitude


class VectorSpaceRandomDataGenerator(RandomDataGenerator):
    def point_to_project(self, n_points=1):
        return self.random_point(n_points)


class MatrixVectorSpaceRandomDataGenerator(VectorSpaceRandomDataGenerator):
    def random_basis_representation(self, n_points=1):
        dtype = (
            gs.get_default_cdtype()
            if isinstance(self, ComplexMatrixVectorSpace)
            else None
        )

        if n_points == 1:
            return gs.random.rand(self.space.dim, dtype=dtype)

        return gs.reshape(
            gs.random.rand(n_points * self.space.dim),
            (n_points, -1),
            dtype=dtype,
        )


class EmbeddedSpaceRandomDataGenerator(RandomDataGenerator):
    def point_to_project(self, n_points=1):
        return self.space.embedding_space.random_point(n_points)


class NFoldManifoldRandomDataGenerator(RandomDataGenerator):
    def point_to_project(self, n_points=1):
        base = self.space.base_manifold
        if not hasattr(base, "embedding_space"):
            raise NotImplementedError("Can't get point to project.")

        n_copies = self.space.n_copies
        point = base.embedding_space.random_point(n_points * n_copies)

        shape = (n_points, n_copies) if n_points > 1 else (n_copies,)
        return gs.reshape(point, shape + base.shape)


class HypersphereIntrinsicRandomDataGenerator(RandomDataGenerator):
    def random_tangent_vec(self, base_point):
        n_points = get_n_points(self.space.point_ndim, base_point)
        batch_shape = (n_points,) if n_points > 1 else ()
        return gs.random.uniform(size=batch_shape + (self.space.dim,)) / self.amplitude


class RankKPSDMatricesRandomDataGenerator(RandomDataGenerator):
    def __init__(self, space, amplitude=1.0):
        super().__init__(space, amplitude=amplitude)
        self._square = SquareMatrices(self.space.n)

    def point_to_project(self, n_points=1):
        return self._square.random_point(n_points)


class LieGroupVectorRandomDataGenerator(RandomDataGenerator):
    def __init__(self, space, amplitude=1.0):
        super().__init__(space, amplitude=amplitude)
        self._euclidean = Euclidean(self.space.dim, equip=False)

    def point_to_project(self, n_points=1):
        return self._euclidean.random_point(n_points)


class KendalShapeRandomDataGenerator(EmbeddedSpaceRandomDataGenerator):
    def random_horizontal_vec(self, base_point):
        tangent_vec = self.random_tangent_vec(base_point)
        return (
            self.space.fiber_bundle.horizontal_projection(tangent_vec, base_point)
            / self.amplitude
        )


class GammaRandomDataGenerator(EmbeddedSpaceRandomDataGenerator):
    def random_point_standard(self, n_points=1):
        return self.space.natural_to_standard(self.random_point(n_points=n_points))

    def random_tangent_vec_standard(self, base_point):
        base_point_natural = self.space.standard_to_natural(base_point)
        tangent_vec_natural = self.random_tangent_vec(base_point_natural)
        return (
            self.space.tangent_natural_to_standard(
                tangent_vec_natural, base_point_natural
            )
            / self.amplitude
        )


class HeisenbergVectorsRandomDataGenerator(VectorSpaceRandomDataGenerator):
    def random_upper_triangular_matrix(self, n_points=1):
        if n_points == 1:
            size = 3
            expected_shape = (3, 3)
            indices = [(0, 1), (0, 2), (1, 2)]
        else:
            size = n_points * 3
            expected_shape = (n_points, 3, 3)
            indices = []
            for k in range(n_points):
                indices.extend([(k, 0, 1), (k, 0, 2), (k, 1, 2)])

        vec = gs.random.uniform(size=size)
        return gs.array_from_sparse(indices, vec, expected_shape) + gs.eye(3)


class ShapeBundleRandomDataGenerator(RandomDataGenerator):
    def __init__(self, space, n_discretized_curves=5):
        super().__init__(space)
        self.n_discretized_curves = n_discretized_curves

    def random_tangent_vec(self, base_point):
        n_points = base_point.shape[0] if base_point.ndim > 2 else 1
        point = self.random_point(n_points=n_points)

        geo = self.space.metric.geodesic(initial_point=base_point, end_point=point)

        times = gs.linspace(0.0, 1.0, self.n_discretized_curves)

        geod = geo(times)

        return (
            self.n_discretized_curves * (geod[..., 1, :, :] - geod[..., 0, :, :])
        ) / self.amplitude



import random

import numpy as np
import pytest
import scipy

import geomstats.backend as gs
from geomstats.test.test_case import TestCase
from geomstats.test.vectorization import generate_vectorization_data

# TODO: use new strategy for vectorization tests


def _convert_gs_to_np(value):
    if gs.is_array(value):
        return gs.to_numpy(value)

    if isinstance(value, (list, tuple)):
        new_value = []
        for value_ in value:
            new_value.append(_convert_gs_to_np(value_))

        if isinstance(value, tuple):
            new_value = tuple(new_value)

        return new_value

    if isinstance(value, dict):
        return {key: _convert_gs_to_np(value_) for key, value_ in value.items()}

    return value


def convert_gs_to_np(*args):
    out = []
    for arg in args:
        val = _convert_gs_to_np(arg)
        out.append(val)

    if len(args) == 1:
        return out[0]

    return out


def get_backend_fncs(func_name, cmp_package=np):
    func_name_ls = func_name.split(".")
    gs_func = gs
    cmp_func = cmp_package
    for name in func_name_ls:
        gs_func = getattr(gs_func, name)
        cmp_func = getattr(cmp_func, name)

    return gs_func, cmp_func


def get_backend_fnc(func_name):
    func_name_ls = func_name.split(".")
    gs_func = gs
    for name in func_name_ls:
        gs_func = getattr(gs_func, name)

    return gs_func


def _get_n_reps(n_reps):
    if n_reps is None:
        return random.randint(2, 5)

    return n_reps


class _AgainstNumpyBasedTestCase(TestCase):
    def _test_func_np_based(self, package, func_name, args):
        gs_fnc, np_fnc = get_backend_fncs(func_name, cmp_package=package)
        np_args = convert_gs_to_np(args)

        gs_array = gs_fnc(*args)
        np_array = np_fnc(*np_args)

        self.assertAllCloseToNp(gs_array, np_array)


class AgainstNumpyTestCase(_AgainstNumpyBasedTestCase):
    def test_array_like_np(self, func_name, args):
        return self.test_func_like_np(func_name, args)

    def test_func_like_np(self, func_name, args):
        return self._test_func_np_based(np, func_name, args)

    def test_unary_op_like_np(self, func_name, a):
        return self.test_func_like_np(func_name, [a])

    def test_binary_op_like_np(self, func_name, a, b):
        return self.test_func_like_np(func_name, [a, b])


class AgainstScipyTestCase(_AgainstNumpyBasedTestCase):
    def _test_func_like_scipy(self, func_name, args):
        return self._test_func_np_based(scipy, func_name, args)

    def test_unary_op_like_scipy(self, func_name, a):
        return self._test_func_like_scipy(func_name, [a])


class AgainstEinsumTestCase(TestCase):
    def test_binary_op_like_einsum(self, func_name, a, b, einsum_expr):
        gs_fnc = get_backend_fnc(func_name)

        gs_out = gs_fnc(a, b)
        ein_out = gs.einsum(einsum_expr, a, b)

        self.assertAllClose(gs_out, ein_out)


class BackendTestCase(TestCase):
    def test_compose_with_inverse(self, func_name_1, func_name_2, a):
        gs_fnc_1 = get_backend_fnc(func_name_1)
        gs_fnc_2 = get_backend_fnc(func_name_2)

        out = gs_fnc_2(gs_fnc_1(a))
        self.assertAllClose(out, a)

    def test_unary_op(self, func_name, a, expected, atol):
        gs_fnc = get_backend_fnc(func_name)
        res = gs_fnc(a)
        self.assertAllClose(res, expected, atol)

    @pytest.mark.vec
    def test_unary_op_vec(self, func_name, a, atol, n_reps=None):
        n_reps = _get_n_reps(n_reps)
        gs_fnc = get_backend_fnc(func_name)
        expected = gs_fnc(a)

        vec_data = generate_vectorization_data(
            data=[dict(func_name=func_name, a=a, expected=expected, atol=atol)],
            arg_names=["a"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    def test_binary_op(self, func_name, a, b, expected, atol):
        gs_fnc = get_backend_fnc(func_name)
        res = gs_fnc(a, b)
        self.assertAllClose(res, expected, atol)

    @pytest.mark.vec
    def test_binary_op_vec(self, func_name, a, b, atol, n_reps=None):
        n_reps = _get_n_reps(n_reps)
        gs_fnc = get_backend_fnc(func_name)
        expected = gs_fnc(a, b)

        vec_data = generate_vectorization_data(
            data=[dict(func_name=func_name, a=a, b=b, expected=expected, atol=atol)],
            arg_names=["a", "b"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    def test_binary_op_raises_error(self, func_name, a, b):
        gs_fnc = get_backend_fnc(func_name)

        with pytest.raises(Exception):
            gs_fnc(a, b)

    @pytest.mark.shape
    def test_func_out_shape(self, func_name, args, expected):
        gs_fnc = get_backend_fnc(func_name)

        # TODO: better comparison (for more info when failing)
        out = gs_fnc(*args)
        self.assertTrue(gs.shape(out) == expected)

    @pytest.mark.type
    def test_func_out_type(self, func_name, args, expected):
        gs_fnc = get_backend_fnc(func_name)

        out = gs_fnc(*args)
        self.assertTrue(isinstance(out, expected))

    def test_func_out_bool(self, func_name, args, expected):
        gs_fnc = get_backend_fnc(func_name)

        out = gs_fnc(*args)
        if expected:
            self.assertTrue(out)
        else:
            self.assertFalse(out)

    def test_func_out_allclose(self, func_name, expected, args=(), kwargs=None):
        kwargs = kwargs or {}
        gs_fnc = get_backend_fnc(func_name)

        out = gs_fnc(*args, **kwargs)
        self.assertAllClose(out, expected)

    def test_func_out_equal(self, func_name, args, expected):
        gs_fnc = get_backend_fnc(func_name)

        out = gs_fnc(*args)
        self.assertEqual(out, expected)


@pytest.mark.type
class DtypeTestCase(TestCase):
    dtypes_str = ["float32", "float64"]  # sort by wider
    default_dtype = gs.as_dtype("float64")

    def _reset_dtype(self):
        gs.set_default_dtype("float64")

    def assertDtype(self, actual, expected):
        self._reset_dtype()

        msg = f"{actual} instead of {expected}"
        self.assertTrue(actual == expected, msg)

    def test_array(self, ls, global_dtype_str, expected_dtype):
        gs.set_default_dtype(global_dtype_str)

        out = gs.array(ls)
        self.assertDtype(out.dtype, expected_dtype)

    def test_array_creation(self, func_name, args, kwargs):
        gs_fnc = get_backend_fnc(func_name)

        for dtype_str in self.dtypes_str:
            dtype = gs.set_default_dtype(dtype_str)
            out = gs_fnc(*args, **kwargs)
            self.assertDtype(out.dtype, dtype)

    def test_array_creation_with_dtype(self, func_name, args=(), kwargs=None):
        kwargs = kwargs or {}
        gs_fnc = get_backend_fnc(func_name)

        for dtype_str in self.dtypes_str:
            # test global
            dtype = gs.set_default_dtype(dtype_str)
            out = gs_fnc(*args, **kwargs)
            self.assertDtype(out.dtype, dtype)

            # test specifying dtype
            for dtype_inner_str in self.dtypes_str:
                dtype_inner = gs.as_dtype(dtype_inner_str)
                out = gs_fnc(*args, dtype=dtype_inner, **kwargs)
                self.assertDtype(out.dtype, dtype_inner)

    def test_array_creation_with_dtype_from_shape(self, func_name, shape):
        return self.test_array_creation_with_dtype(func_name, (shape,), {})

    def test_unary_op_float_input(self, func_name, x):
        gs_fnc = get_backend_fnc(func_name)

        for dtype_str in self.dtypes_str:
            dtype = gs.set_default_dtype(dtype_str)
            out = gs_fnc(x)

            self.assertDtype(out.dtype, dtype)

    def test_array_creation_with_dtype_from_array(
        self,
        func_name,
        array_shape,
        kwargs=None,
        func_array=gs.ones,
    ):
        self._reset_dtype()

        kwargs = kwargs or {}
        gs_fnc = get_backend_fnc(func_name)

        for dtype_str in self.dtypes_str:
            # test dynamic
            dtype = gs.as_dtype(dtype_str)
            a = func_array(array_shape, dtype=dtype)

            out = gs_fnc(a, **kwargs)
            self.assertDtype(out.dtype, a.dtype)

            # test specifying dtype
            for dtype_inner_str in self.dtypes_str:
                dtype_inner = gs.as_dtype(dtype_inner_str)
                out = gs_fnc(a, dtype=dtype_inner, **kwargs)
                self.assertDtype(out.dtype, dtype_inner)

    def test_from_numpy_given_shape(
        self, array_shape, kwargs=None, np_func_array=gs.ones
    ):
        self._reset_dtype()

        func_name = "from_numpy"
        kwargs = kwargs or {}
        gs_fnc = get_backend_fnc(func_name)

        for dtype_str in self.dtypes_str:
            # test dynamic
            np_dtype = np.dtype(dtype_str)
            dtype = gs.as_dtype(dtype_str)
            a = np_func_array(array_shape, dtype=np_dtype)

            out = gs_fnc(a, **kwargs)
            self.assertDtype(out.dtype, dtype)

    def test_to_numpy_given_shape(self, array_shape, kwargs=None, func_array=gs.ones):
        self._reset_dtype()

        func_name = "to_numpy"

        kwargs = kwargs or {}
        gs_fnc = get_backend_fnc(func_name)

        for dtype_str in self.dtypes_str:
            np_dtype = np.dtype(dtype_str)
            dtype = gs.as_dtype(dtype_str)
            a = func_array(array_shape, dtype=dtype)

            out = gs_fnc(a, **kwargs)
            self.assertDtype(out.dtype, np_dtype)

    def test_unary_op_with_dtype_given_shape(self, func_name, array_shape, kwargs=None):
        return self.test_array_creation_with_dtype_from_array(
            func_name, array_shape, kwargs=kwargs
        )

    def test_unary_op_given_shape(
        self, func_name, array_shape, kwargs=None, func_array=gs.ones
    ):
        self._reset_dtype()

        kwargs = kwargs or {}
        gs_fnc = get_backend_fnc(func_name)

        for dtype_str in self.dtypes_str:
            dtype = gs.as_dtype(dtype_str)
            a = func_array(array_shape, dtype=dtype)

            out = gs_fnc(a, **kwargs)
            self.assertDtype(out.dtype, dtype)

    def test_unary_op_mult_out_given_shape(
        self, func_name, array_shape, func_array=gs.ones
    ):
        self._reset_dtype()

        gs_fnc = get_backend_fnc(func_name)

        for dtype_str in self.dtypes_str:
            dtype = gs.as_dtype(dtype_str)
            a = func_array(array_shape, dtype=dtype)

            out = gs_fnc(a)
            for out_ in out:
                self.assertDtype(out_.dtype, dtype)

    def _test_op_given_array(self, func_name, create_array):
        # create_array to avoid using cast
        gs_fnc = get_backend_fnc(func_name)

        for dtype_str in self.dtypes_str:
            dtype = gs.set_default_dtype(dtype_str)
            args = create_array()
            if gs.is_array(args):
                args = [args]

            out = gs_fnc(*args)
            self.assertDtype(out.dtype, dtype)

    def test_unary_op_given_array(self, func_name, create_array):
        return self._test_op_given_array(func_name, create_array)

    def test_binary_op_float_input(self, func_name, x1, x2):
        gs_fnc = get_backend_fnc(func_name)

        for dtype_str in self.dtypes_str:
            dtype = gs.set_default_dtype(dtype_str)
            out = gs_fnc(x1, x2)

            self.assertDtype(out.dtype, dtype)

    def test_binary_op_given_shape(
        self, func_name, shape_a, shape_b, kwargs=None, func_a=gs.ones, func_b=gs.ones
    ):
        self._reset_dtype()

        kwargs = kwargs or {}
        gs_fnc = get_backend_fnc(func_name)

        for i, dtype_a_str in enumerate(self.dtypes_str):
            dtype_a = gs.as_dtype(dtype_a_str)

            a = func_a(shape_a, dtype=dtype_a)

            for j, dtype_b_str in enumerate(self.dtypes_str):
                dtype_b = gs.as_dtype(dtype_b_str)

                b = func_b(shape_b, dtype=dtype_b)

                out = gs_fnc(a, b, **kwargs)
                cmp_dtype = dtype_a if i > j else dtype_b

                self.assertDtype(out.dtype, cmp_dtype)

    def test_ternary_op_given_shape(
        self,
        func_name,
        shape_a,
        shape_b,
        shape_c,
        func_a=gs.ones,
        func_b=gs.ones,
        func_c=gs.ones,
    ):
        self._reset_dtype()

        gs_fnc = get_backend_fnc(func_name)

        for i, dtype_a_str in enumerate(self.dtypes_str):
            dtype_a = gs.as_dtype(dtype_a_str)

            a = func_a(shape_a, dtype=dtype_a)

            for j, dtype_b_str in enumerate(self.dtypes_str):
                dtype_b = gs.as_dtype(dtype_b_str)

                b = func_b(shape_b, dtype=dtype_b)

                for k, dtype_c_str in enumerate(self.dtypes_str):
                    dtype_c = gs.as_dtype(dtype_c_str)
                    c = func_c(shape_c, dtype=dtype_c)

                    out = gs_fnc(a, b, c)
                    cmp_dtype = gs.as_dtype(self.dtypes_str[max([i, j, k])])

                    self.assertDtype(out.dtype, cmp_dtype)

    def test_ternary_op_given_array(self, func_name, create_array):
        return self._test_op_given_array(func_name, create_array)

    def test_func_out_dtype(self, func_name, args, kwargs, expected):
        self._reset_dtype()

        gs_fnc = get_backend_fnc(func_name)

        out = gs_fnc(*args, **kwargs)

        self.assertDtype(out.dtype, expected)

    def test_solve_sylvester(
        self,
        shape_a,
        shape_b,
        shape_c,
        func_a=gs.ones,
        func_b=gs.ones,
        func_c=gs.ones,
    ):
        func_name = "linalg.solve_sylvester"

        self.test_ternary_op_given_shape(
            func_name,
            shape_a,
            shape_b,
            shape_c,
            func_a=func_a,
            func_b=func_b,
            func_c=func_c,
        )

    def test_random_distrib_complex(self, func_name, args=(), kwargs=None):
        self._reset_dtype()
        kwargs = kwargs or {}

        gs_fnc = get_backend_fnc(func_name)

        dtype = gs.as_dtype("complex128")
        out = gs_fnc(*args, **kwargs, dtype=dtype)

        self.assertDtype(out.dtype, dtype)
        self.assertTrue(
            gs.sum(gs.abs(gs.imag(out))) > gs.atol, msg="Zero imaginary part"
        )



"""Abstract classes for manifolds.

Lead authors: Nicolas Guigui and Nina Miolane.
"""

import abc
import math

import geomstats.backend as gs
from geomstats.geometry.complex_manifold import ComplexManifold
from geomstats.geometry.manifold import Manifold
from geomstats.geometry.pullback_metric import PullbackMetric
from geomstats.vectorization import get_batch_shape


class VectorSpace(Manifold, abc.ABC):
    """Abstract class for vector spaces.

    Parameters
    ----------
    shape : tuple
        Shape of the elements of the vector space. The dimension is the
        product of these values by default.
    """

    def __init__(self, shape, dim=None, **kwargs):
        if dim is None:
            dim = math.prod(shape)
        super().__init__(dim=dim, shape=shape, **kwargs)
        self._basis = None

    def belongs(self, point, atol=gs.atol):
        """Evaluate if the point belongs to the vector space.

        This method checks the shape of the input point.

        Parameters
        ----------
        point : array-like, shape=[.., *point_shape]
            Point to test.
        atol : float
            Unused here.

        Returns
        -------
        belongs : array-like, shape=[...,]
            Boolean evaluating if point belongs to the space.
        """
        belongs = self.shape == point.shape[-self.point_ndim :]
        shape = point.shape[: -self.point_ndim]
        if belongs:
            return gs.ones(shape, dtype=bool)
        return gs.zeros(shape, dtype=bool)

    def is_tangent(self, vector, base_point=None, atol=gs.atol):
        """Check whether the vector is tangent at base_point.

        Tangent vectors are identified with points of the vector space so
        this checks the shape of the input vector.

        Parameters
        ----------
        vector : array-like, shape=[..., *point_shape]
            Vector.
        base_point : array-like, shape=[..., *point_shape]
            Point in the vector space.
        atol : float
            Absolute tolerance.
            Optional, default: backend atol.

        Returns
        -------
        is_tangent : array-like, shape=[...,]
            Boolean denoting if vector is a tangent vector at the base point.
        """
        belongs = self.belongs(vector, atol)
        if base_point is not None and base_point.ndim > vector.ndim:
            return gs.broadcast_to(belongs, base_point.shape[: -self.point_ndim])
        return belongs

    def to_tangent(self, vector, base_point=None):
        """Project a vector to a tangent space of the vector space.

        This method is for compatibility and returns vector.

        Parameters
        ----------
        vector : array-like, shape=[..., *point_shape]
            Vector.
        base_point : array-like, shape=[..., *point_shape]
            Point in the vector space

        Returns
        -------
        tangent_vec : array-like, shape=[..., *point_shape]
            Tangent vector at base point.
        """
        tangent_vec = self.projection(vector)
        if base_point is not None and base_point.ndim > vector.ndim:
            return gs.broadcast_to(tangent_vec, base_point.shape)
        return tangent_vec

    def random_point(self, n_samples=1, bound=1.0):
        """Sample in the vector space with a uniform distribution in a box.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.
        bound : float
            Side of hypercube support of the uniform distribution.
            Optional, default: 1.0

        Returns
        -------
        point : array-like, shape=[..., dim]
           Sample.
        """
        size = (self.dim,)
        if n_samples != 1:
            size = (n_samples,) + size
        return bound * (gs.random.rand(*size) - 0.5) * 2

    def random_tangent_vec(self, base_point=None, n_samples=1):
        """Generate random tangent vec.

        This method is not recommended for statistical purposes, as the
        tangent vectors generated are not drawn from a distribution related
        to the Riemannian metric.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.
        base_point :  array-like, shape={[n_samples, *point_shape], [*point_shape,]}
            Point.

        Returns
        -------
        tangent_vec : array-like, shape=[..., *point_shape]
            Tangent vec at base point.
        """
        if base_point is None:
            return self.random_point(n_samples)

        if (
            n_samples > 1
            and base_point.ndim > self.point_ndim
            and n_samples != base_point.shape[0]
        ):
            raise ValueError(
                "The number of base points must be the same as the "
                "number of samples, when the number of base points is different from 1."
            )
        if n_samples == 1 and base_point.ndim > self.point_ndim:
            n_samples = base_point.shape[0]
        return self.random_point(n_samples)

    @property
    def basis(self):
        """Basis of the vector space."""
        if self._basis is None:
            self._basis = self._create_basis()
        return self._basis

    @abc.abstractmethod
    def _create_basis(self):
        """Create a canonical basis."""


class MatrixVectorSpace(VectorSpace, abc.ABC):
    """A matrix vector space."""

    @abc.abstractmethod
    def basis_representation(self, matrix_representation):
        """Compute the coefficients of matrices in the given basis.

        This takes a matrix (the matrix representation of a point) and
        transforms it into its corresponding vector representation
        (the coefficients wrt a given basis).

        Previously, this method was called `to_vector`. `basis_representation`
        makes it more clear that the vector representation depends on the chosen
        basis.

        Parameters
        ----------
        matrix_representation : array-like, shape=[..., *point_shape]
            Matrix.

        Returns
        -------
        basis_representation : array-like, shape=[..., dim]
            Coefficients in the basis.
        """
        raise NotImplementedError("basis_representation not implemented.")

    def matrix_representation(self, basis_representation):
        """Compute the matrix representation for the given basis coefficients.

        This takes a vector representation of a point (the coefficients wrt
        a given basis) and creates the corresponding matrix representation.

        Parameters
        ----------
        basis_representation : array-like, shape=[..., dim]
            Coefficients in the basis.

        Returns
        -------
        matrix_representation : array-like, shape=[..., *point_shape]
            Matrix.
        """
        return gs.einsum("...i,ijk ->...jk", basis_representation, self.basis)

    def random_point(self, n_samples=1, bound=1.0):
        """Sample in the vector space with a uniform distribution in a box.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.
        bound : float
            Side of hypercube support of the uniform distribution.
            Optional, default: 1.0

        Returns
        -------
        point : array-like, shape=[..., *point_shape]
           Sample.
        """
        return self.matrix_representation(super().random_point(n_samples, bound))


class ComplexVectorSpace(ComplexManifold, abc.ABC):
    """Abstract class for complex vector spaces.

    Parameters
    ----------
    shape : tuple
        Shape of the elements of the vector space. The dimension is the
        product of these values by default.
    """

    def __init__(self, shape, dim=None, **kwargs):
        if dim is None:
            dim = math.prod(shape)
        super().__init__(shape=shape, dim=dim, **kwargs)

    def belongs(self, point, atol=gs.atol):
        """Evaluate if the point belongs to the vector space.

        This method checks the shape of the input point.

        Parameters
        ----------
        point : array-like, shape=[.., *point_shape]
            Point to test.
        atol : float
            Unused here.

        Returns
        -------
        belongs : array-like, shape=[...,]
            Boolean evaluating if point belongs to the space.
        """
        belongs = self.shape == point.shape[-self.point_ndim :]
        shape = point.shape[: -self.point_ndim]
        if belongs:
            return gs.ones(shape, dtype=bool)
        return gs.zeros(shape, dtype=bool)

    def is_tangent(self, vector, base_point=None, atol=gs.atol):
        """Check whether the vector is tangent at base_point.

        Tangent vectors are identified with points of the vector space so
        this checks the shape of the input vector.

        Parameters
        ----------
        vector : array-like, shape=[..., *point_shape]
            Vector.
        base_point : array-like, shape=[..., *point_shape]
            Point in the vector space.
        atol : float
            Absolute tolerance.
            Optional, default: backend atol.

        Returns
        -------
        is_tangent : bool
            Boolean denoting if vector is a tangent vector at the base point.
        """
        belongs = self.belongs(vector, atol)
        if base_point is not None and base_point.ndim > vector.ndim:
            return gs.broadcast_to(belongs, base_point.shape[: -self.point_ndim])
        return belongs

    def to_tangent(self, vector, base_point=None):
        """Project a vector to a tangent space of the vector space.

        This method is for compatibility and returns vector.

        Parameters
        ----------
        vector : array-like, shape=[..., *point_shape]
            Vector.
        base_point : array-like, shape=[..., *point_shape]
            Point in the vector space

        Returns
        -------
        tangent_vec : array-like, shape=[..., *point_shape]
            Tangent vector at base point.
        """
        tangent_vec = self.projection(vector)
        if base_point is not None and base_point.ndim > vector.ndim:
            return gs.broadcast_to(tangent_vec, base_point.shape)
        return tangent_vec

    def random_point(self, n_samples=1, bound=1.0):
        """Sample in the complex vector space with a uniform distribution in a box.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.
        bound : float
            Side of hypercube support of the uniform distribution.
            Optional, default: 1.0

        Returns
        -------
        point : array-like, shape=[..., *point_shape]
           Sample.
        """
        size = self.shape
        if n_samples != 1:
            size = (n_samples,) + self.shape
        point = bound * (
            gs.random.rand(*size, dtype=gs.get_default_cdtype()) - 0.5 - 0.5j
        )
        return point


class ComplexMatrixVectorSpace(ComplexVectorSpace):
    """A matrix vector space."""


class LevelSet(Manifold, abc.ABC):
    """Class for manifolds embedded in a vector space by a submersion.

    Parameters
    ----------
    intrinsic : bool
        Coordinates type.
    """

    def __init__(self, intrinsic=False, shape=None, **kwargs):
        self.embedding_space = self._define_embedding_space()

        if shape is None:
            shape = self.embedding_space.shape

        super().__init__(intrinsic=intrinsic, shape=shape, **kwargs)

    @abc.abstractmethod
    def _define_embedding_space(self):
        """Define embedding space of the manifold.

        Returns
        -------
        embedding_space : Manifold
            Instance of Manifold.
        """

    @abc.abstractmethod
    def submersion(self, point):
        r"""Submersion that defines the manifold.

        :math:`\mathrm{submersion}(x)=0` defines the manifold.

        Parameters
        ----------
        point : array-like, shape=[..., *point_shape]

        Returns
        -------
        submersed_point : array-like
        """

    @abc.abstractmethod
    def tangent_submersion(self, vector, point):
        """Tangent submersion.

        Parameters
        ----------
        vector : array-like, shape=[..., *point_shape]
        point : array-like, shape=[..., *point_shape]

        Returns
        -------
        submersed_vector : array-like
        """

    def belongs(self, point, atol=gs.atol):
        """Evaluate if a point belongs to the manifold.

        Parameters
        ----------
        point : array-like, shape=[..., *point_shape]
            Point to evaluate.
        atol : float
            Absolute tolerance.
            Optional, default: backend atol.

        Returns
        -------
        belongs : array-like, shape=[...,]
            Boolean evaluating if point belongs to the manifold.
        """
        belongs = self.embedding_space.belongs(point, atol)
        if not gs.any(belongs):
            return belongs

        submersed_point = self.submersion(point)

        n_batch = gs.ndim(point) - len(self.shape)
        axis = tuple(range(-len(submersed_point.shape) + n_batch, 0))

        if gs.is_complex(submersed_point):
            constraint = gs.isclose(submersed_point, 0.0 + 0.0j, atol=atol)
        else:
            constraint = gs.isclose(submersed_point, 0.0, atol=atol)

        if axis:
            constraint = gs.all(constraint, axis=axis)

        return gs.logical_and(belongs, constraint)

    def is_tangent(self, vector, base_point, atol=gs.atol):
        """Check whether the vector is tangent at base_point.

        Parameters
        ----------
        vector : array-like, shape=[..., *point_shape]
            Vector.
        base_point : array-like, shape=[..., *point_shape]
            Point on the manifold.
        atol : float
            Absolute tolerance.
            Optional, default: backend atol.

        Returns
        -------
        is_tangent : bool
            Boolean denoting if vector is a tangent vector at the base point.
        """
        belongs = self.embedding_space.is_tangent(vector, base_point, atol)
        if not gs.any(belongs):
            return belongs

        submersed_vector = self.tangent_submersion(vector, base_point)

        n_batch = len(get_batch_shape(self.point_ndim, base_point, vector))
        axis = tuple(range(-len(submersed_vector.shape) + n_batch, 0))

        constraint = gs.isclose(submersed_vector, 0.0, atol=atol)
        if axis:
            constraint = gs.all(constraint, axis=axis)

        return gs.logical_and(belongs, constraint)

    def intrinsic_to_extrinsic_coords(self, point_intrinsic):
        """Convert from intrinsic to extrinsic coordinates.

        Parameters
        ----------
        point_intrinsic : array-like, shape=[..., *point_shape]
            Point in the embedded manifold in intrinsic coordinates.

        Returns
        -------
        point_extrinsic : array-like, shape=[..., *embedding_space.point_shape]
            Point in the embedded manifold in extrinsic coordinates.
        """
        raise NotImplementedError("intrinsic_to_extrinsic_coords is not implemented.")

    def extrinsic_to_intrinsic_coords(self, point_extrinsic):
        """Convert from extrinsic to intrinsic coordinates.

        Parameters
        ----------
        point_extrinsic : array-like, shape=[..., *embedding_space.point_shape]
            Point in the embedded manifold in extrinsic coordinates,
            i. e. in the coordinates of the embedding manifold.

        Returns
        -------
        point_intrinsic : array-lie, shape=[..., *point_shape]
            Point in the embedded manifold in intrinsic coordinates.
        """
        raise NotImplementedError("extrinsic_to_intrinsic_coords is not implemented.")


class OpenSet(Manifold, abc.ABC):
    """Class for manifolds that are open sets.

    NB: if the embedding space is a vector space, use `VectorSpaceOpenSet`.

    Parameters
    ----------
    embedding_space: Manifold
        Embedding space that contains the manifold.
    """

    def __init__(self, embedding_space, shape=None, **kwargs):
        self.embedding_space = embedding_space
        if shape is None:
            shape = embedding_space.shape
        super().__init__(shape=shape, **kwargs)

    def is_tangent(self, vector, base_point=None, atol=gs.atol):
        """Check whether the vector is tangent at base_point.

        Parameters
        ----------
        vector : array-like, shape=[..., *point_shape]
            Vector.
        base_point : array-like, shape=[..., *point_shape]
            Point on the manifold.
        atol : float
            Absolute tolerance.
            Optional, default: backend atol.

        Returns
        -------
        is_tangent : bool
            Boolean denoting if vector is a tangent vector at the base point.
        """
        return self.embedding_space.is_tangent(vector, base_point, atol)

    def to_tangent(self, vector, base_point=None):
        """Project a vector to a tangent space of the manifold.

        Parameters
        ----------
        vector : array-like, shape=[..., *point_shape]
            Vector.
        base_point : array-like, shape=[..., *point_shape]
            Point on the manifold.

        Returns
        -------
        tangent_vec : array-like, shape=[..., *point_shape]
            Tangent vector at base point.
        """
        return self.embedding_space.to_tangent(vector, base_point)

    def random_point(self, n_samples=1, bound=1.0):
        """Sample random points on the manifold.

        Points are sampled from the embedding space using the distribution set
        for that manifold and then projected to the manifold. As a result, this
        is not a uniform distribution on the manifold itself.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.
        bound : float
            Bound of the interval in which to sample for the embedding space.
            Optional, default: 1.

        Returns
        -------
        samples : array-like, shape=[..., *point_shape]
            Points sampled on the hypersphere.
        """
        sample = self.embedding_space.random_point(n_samples, bound)
        return self.projection(sample)


class VectorSpaceOpenSet(OpenSet, abc.ABC):
    """Class for manifolds that are open sets of a vector space.

    In this case, tangent vectors are identified with vectors of the embedding
    space.

    Parameters
    ----------
    embedding_space: VectorSpace
        Embedding space that contains the manifold.
    """

    def is_tangent(self, vector, base_point=None, atol=gs.atol):
        """Check whether the vector is tangent at base_point.

        Parameters
        ----------
        vector : array-like, shape=[..., *point_shape]
            Vector.
        base_point : array-like, shape=[..., *point_shape]
            Point on the manifold.
        atol : float
            Absolute tolerance.
            Optional, default: backend atol.

        Returns
        -------
        is_tangent : bool
            Boolean denoting if vector is a tangent vector at the base point.
        """
        is_tangent = self.embedding_space.belongs(vector, atol)
        if base_point is not None and base_point.ndim > vector.ndim:
            return gs.broadcast_to(is_tangent, base_point.shape[: -self.point_ndim])
        return is_tangent

    def to_tangent(self, vector, base_point=None):
        """Project a vector to a tangent space of the manifold.

        Parameters
        ----------
        vector : array-like, shape=[..., *point_shape]
            Vector.
        base_point : array-like, shape=[..., *point_shape]
            Point on the manifold.

        Returns
        -------
        tangent_vec : array-like, shape=[..., *point_shape]
            Tangent vector at base point.
        """
        tangent_vec = self.embedding_space.projection(vector)
        if base_point is not None and base_point.ndim > vector.ndim:
            return gs.broadcast_to(tangent_vec, base_point.shape)
        return tangent_vec


class ComplexVectorSpaceOpenSet(ComplexManifold, abc.ABC):
    """Class for manifolds that are open sets of a complex vector space.

    In this case, tangent vectors are identified with vectors of the embedding
    space.

    Parameters
    ----------
    dim: int
        Dimension of the manifold. It is often the same as the embedding space
        dimension but may differ in some cases.
    embedding_space: VectorSpace
        Embedding space that contains the manifold.
    """

    def __init__(self, embedding_space, shape=None, **kwargs):
        if shape is None:
            shape = embedding_space.shape
        super().__init__(shape=shape, intrinsic=False, **kwargs)
        self.embedding_space = embedding_space

    def is_tangent(self, vector, base_point=None, atol=gs.atol):
        """Check whether the vector is tangent at base_point.

        Parameters
        ----------
        vector : array-like, shape=[..., *point_shape]
            Vector.
        base_point : array-like, shape=[..., *point_shape]
            Point on the manifold.
        atol : float
            Absolute tolerance.
            Optional, default: backend atol.

        Returns
        -------
        is_tangent : bool
            Boolean denoting if vector is a tangent vector at the base point.
        """
        is_tangent = self.embedding_space.belongs(vector, atol)
        if base_point is not None and base_point.ndim > vector.ndim:
            return gs.broadcast_to(is_tangent, base_point.shape[: -self.point_ndim])
        return is_tangent

    def to_tangent(self, vector, base_point=None):
        """Project a vector to a tangent space of the manifold.

        Parameters
        ----------
        vector : array-like, shape=[..., *point_shape]
            Vector.
        base_point : array-like, shape=[..., *point_shape]
            Point on the manifold.

        Returns
        -------
        tangent_vec : array-like, shape=[..., *point_shape]
            Tangent vector at base point.
        """
        tangent_vec = self.embedding_space.projection(vector)
        if base_point is not None and base_point.ndim > vector.ndim:
            return gs.broadcast_to(tangent_vec, base_point.shape)
        return tangent_vec

    def random_point(self, n_samples=1, bound=1.0):
        """Sample random points on the manifold.

        If the manifold is compact, a uniform distribution is used.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.
        bound : float
            Bound of the interval in which to sample for non compact manifolds.
            Optional, default: 1.

        Returns
        -------
        samples : array-like, shape=[..., *point_shape]
            Points sampled on the hypersphere.
        """
        sample = self.embedding_space.random_point(n_samples, bound)
        return self.projection(sample)


class ImmersedSet(Manifold, abc.ABC):
    """Class for manifolds embedded in a vector space by an immersion.

    The manifold is represented with intrinsic coordinates, such that
    the immersion gives a parameterization of the manifold in these
    coordinates.

    Parameters
    ----------
    dim : int
        Dimension of the embedded manifold.
    """

    def __init__(self, dim, equip=True):
        super().__init__(dim=dim, shape=(dim,), intrinsic=True, equip=equip)
        self.embedding_space = self._define_embedding_space()

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return PullbackMetric

    @abc.abstractmethod
    def _define_embedding_space(self):
        """Define embedding space of the manifold.

        Returns
        -------
        embedding_space : Manifold
            Instance of Manifold.
        """

    @abc.abstractmethod
    def immersion(self, point):
        """Evaluate the immersion function at a point.

        Parameters
        ----------
        point : array-like, shape=[..., dim]
            Point in the immersed manifold.

        Returns
        -------
        immersion : array-like, shape=[..., dim_embedding]
            Immersion of the point.
        """

    def tangent_immersion(self, tangent_vec, base_point):
        """Evaluate the tangent immersion at a tangent vec and point.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., dim]
        base_point : array-like, shape=[..., dim]
            Point in the immersed manifold.

        Returns
        -------
        tangent_vec_emb : array-like, shape=[..., dim_embedding]
        """
        jacobian_immersion = self.jacobian_immersion(base_point)
        return gs.matvec(jacobian_immersion, tangent_vec)

    def jacobian_immersion(self, base_point):
        """Evaluate the Jacobian of the immersion at a point.

        Parameters
        ----------
        base_point : array-like, shape=[..., dim]
            Point in the immersed manifold.

        Returns
        -------
        jacobian_immersion : array-like, shape=[..., dim_embedding, dim]
        """
        return gs.autodiff.jacobian_vec(self.immersion)(base_point)

    def hessian_immersion(self, base_point):
        """Compute the Hessian of the immersion.

        Parameters
        ----------
        base_point : array-like, shape=[..., dim]
            Base point.

        Returns
        -------
        hessian_immersion : array-like, shape=[..., embedding_dim, dim, dim]
            Hessian at the base point
        """
        return gs.autodiff.hessian_vec(
            self.immersion, func_out_ndim=self.embedding_space.dim
        )(base_point)

    def is_tangent(self, vector, base_point, atol=gs.atol):
        """Check whether the vector is tangent at base_point.

        Parameters
        ----------
        vector : array-like, shape=[..., dim]
            Vector.
        base_point : array-like, shape=[..., dim]
            Point on the manifold.
        atol : float
            Absolute tolerance.
            Optional, default: backend atol.

        Returns
        -------
        is_tangent : bool
            Boolean denoting if vector is a tangent vector at the base point.
        """
        raise NotImplementedError("`is_tangent` is not implemented yet")

    def belongs(self, point, atol=gs.atol):
        """Evaluate if a point belongs to the manifold.

        Parameters
        ----------
        point : array-like, shape=[..., dim]
            Point to evaluate.
        atol : float
            Absolute tolerance.
            Optional, default: backend atol.

        Returns
        -------
        belongs : array-like, shape=[...,]
            Boolean evaluating if point belongs to the manifold.
        """
        raise NotImplementedError("`is_tangent` is not implemented yet")

    def projection(self, point):
        """Project a point to the embedded manifold.

        This is simply point, since we are in intrinsic coordinates.

        Parameters
        ----------
        point : array-like, shape=[..., dim_embedding]
            Point in the embedding manifold.

        Returns
        -------
        projected_point : array-like, shape=[..., dim]
            Point in the embedded manifold.
        """
        raise NotImplementedError("`projection` is not implemented yet")

    def to_tangent(self, vector, base_point):
        """Project a vector to a tangent space of the manifold.

        This is simply the vector since we are in intrinsic coordinates.

        Parameters
        ----------
        vector : array-like, shape=[..., dim_embedding]
            Vector.
        base_point : array-like, shape=[..., dim]
            Point in the embedded manifold.

        Returns
        -------
        tangent_vec : array-like, shape=[..., dim]
            Tangent vector at base point.
        """
        raise NotImplementedError("`to_tangent` is not implemented yet")

    def random_point(self, n_samples=1, bound=1.0):
        """Sample random points on the manifold according to some distribution.

        If the manifold is compact, preferably a uniform distribution will be used.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.
        bound : float
            Bound of the interval in which to sample for non compact manifolds.
            Optional, default: 1.

        Returns
        -------
        samples : array-like, shape=[..., *point_shape]
            Points sampled on the manifold.
        """
        raise NotImplementedError("`random_point` is not implemented yet")


class DiffeomorphicManifold(Manifold):
    """A manifold defined by a diffeomorphism."""

    def __init__(self, diffeo, image_space, **kwargs):
        self.diffeo = diffeo
        self.image_space = image_space
        super().__init__(**kwargs)

    def belongs(self, point, atol=gs.atol):
        """Evaluate if a point belongs to the manifold.

        Parameters
        ----------
        point : array-like, shape=[..., *point_shape]
            Point to evaluate.
        atol : float
            Absolute tolerance.
            Optional, default: backend atol.

        Returns
        -------
        belongs : array-like, shape=[...,]
            Boolean evaluating if point belongs to the manifold.
        """
        if not self.intrinsic:
            raise ValueError("`belongs` is not implemented.")
        return self.image_space.belongs(self.diffeo(point), atol=atol)

    def is_tangent(self, vector, base_point=None, atol=gs.atol):
        """Check whether the vector is tangent at base_point.

        Parameters
        ----------
        vector : array-like, shape=[..., *point_shape]
            Vector.
        base_point : array-like, shape=[..., *point_shape]
            Point on the manifold.
        atol : float
            Absolute tolerance.
            Optional, default: backend atol.

        Returns
        -------
        is_tangent : bool
            Boolean denoting if vector is a tangent vector at the base point.
        """
        if not self.intrinsic:
            raise ValueError("`is_tangent` is not implemented.")

        image_point = self.diffeo(base_point)
        image_vector = self.diffeo.tangent(
            vector, base_point=base_point, image_point=image_point
        )
        return self.image_space.is_tangent(image_vector, image_point, atol=atol)

    def to_tangent(self, vector, base_point=None):
        """Project a vector to a tangent space of the manifold.

        Parameters
        ----------
        vector : array-like, shape=[..., *point_shape]
            Vector.
        base_point : array-like, shape=[..., *point_shape]
            Point on the manifold.

        Returns
        -------
        tangent_vec : array-like, shape=[..., *point_shape]
            Tangent vector at base point.
        """
        image_point = self.diffeo(base_point)
        image_vector = self.diffeo.tangent(
            vector, base_point=base_point, image_point=image_point
        )
        image_tangent_vec = self.image_space.to_tangent(image_vector, image_point)
        return self.diffeo.inverse_tangent(
            image_tangent_vec, image_point=image_point, base_point=base_point
        )

    def random_point(self, n_samples=1, **kwargs):
        """Sample random points on the manifold according to some distribution.

        If the manifold is compact, preferably a uniform distribution will be used.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.

        Returns
        -------
        samples : array-like, shape=[..., *point_shape]
            Points sampled on the manifold.
        """
        image_point = self.image_space.random_point(n_samples=n_samples, **kwargs)
        return self.diffeo.inverse(image_point)

    def regularize(self, point):
        """Regularize a point to the canonical representation for the manifold.

        Parameters
        ----------
        point : array-like, shape=[..., dim]
            Point.

        Returns
        -------
        regularized_point : array-like, shape=[..., *point_shape]
            Regularized point.
        """
        image_point = self.diffeo(point)
        regularized_image_point = self.image_space.regularize(image_point)
        return self.diffeo.inverse(regularized_image_point)

    def random_tangent_vec(self, base_point=None, n_samples=1):
        """Generate random tangent vec.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.
        base_point :  array-like, shape={[n_samples, *point_shape], [*point_shape,]}
            Point.

        Returns
        -------
        tangent_vec : array-like, shape=[..., *point_shape]
            Tangent vec at base point.
        """
        image_point = self.diffeo(base_point)
        image_tangent_vec = self.image_space.random_tangent_vec(
            image_point, n_samples=n_samples
        )
        return self.diffeo.inverse_tangent(
            image_tangent_vec, image_point=image_point, base_point=base_point
        )


class DiffeomorphicVectorSpace(VectorSpace, DiffeomorphicManifold):
    """A vector space defined by a diffeomorphism."""

    def projection(self, point):
        r"""Make a matrix null-row-sum symmetric.

        It considers only the first :math:`n-1 \times n-1` components.

        Parameters
        ----------
        point : array-like, shape=[..., n, n]
            Matrix.

        Returns
        -------
        sym : array-like, shape=[..., n, n]
            Symmetric matrix.
        """
        image_point = self.diffeo(point)
        proj_image_point = self.image_space.projection(image_point)
        return self.diffeo.inverse(proj_image_point)


class DiffeomorphicMatrixVectorSpace(MatrixVectorSpace, DiffeomorphicVectorSpace):
    """A matrix vector space defined by a diffeomorphism."""

    def basis_representation(self, matrix_representation):
        """Convert a symmetric matrix into a vector.

        Parameters
        ----------
        matrix_representation : array-like, shape=[..., n, n]
            Matrix.

        Returns
        -------
        basis_representation : array-like, shape=[..., n(n+1)/2]
            Vector.
        """
        image_matrix_representation = self.diffeo(matrix_representation)
        return self.image_space.basis_representation(image_matrix_representation)

    def matrix_representation(self, basis_representation):
        """Convert a vector into a symmetric matrix.

        Parameters
        ----------
        basis_representation : array-like, shape=[..., n(n+1)/2]
            Vector.

        Returns
        -------
        matrix_representation : array-like, shape=[..., n, n]
            Symmetric matrix.
        """
        image_point = self.image_space.matrix_representation(basis_representation)
        return self.diffeo.inverse(image_point)



"""Riemannian and pseudo-Riemannian metrics for complex manifolds.

Lead author: Yann Cabanes.
"""

import geomstats.backend as gs
from geomstats.geometry.riemannian_metric import RiemannianMetric
from geomstats.vectorization import check_is_batch


class ComplexRiemannianMetric(RiemannianMetric):
    r"""Class for Riemannian and pseudo-Riemannian metrics for Complex manifolds.

    The associated Levi-Civita connection on the tangent bundle.
    """

    def inner_product(self, tangent_vec_a, tangent_vec_b, base_point=None):
        """Inner product between two tangent vectors at a base point.

        Parameters
        ----------
        tangent_vec_a: array-like, shape=[..., dim]
            Tangent vector at base point.
        tangent_vec_b: array-like, shape=[..., dim]
            Tangent vector at base point.
        base_point: array-like, shape=[..., dim]
            Base point.
            Optional, default: None.

        Returns
        -------
        inner_product : array-like, shape=[...,]
            Inner-product.
        """
        inner_prod_mat = self.metric_matrix(base_point)
        aux = gs.einsum("...j,...jk->...k", gs.conj(tangent_vec_a), inner_prod_mat)
        return gs.dot(aux, tangent_vec_b)

    def squared_norm(self, vector, base_point=None):
        """Compute the square of the norm of a vector.

        Squared norm of a vector associated to the inner product
        at the tangent space at a base point.

        Parameters
        ----------
        vector : array-like, shape=[..., dim]
            Vector.
        base_point : array-like, shape=[..., dim]
            Base point.

        Returns
        -------
        sq_norm : array-like, shape=[...,]
            Squared norm.
        """
        sq_norm = self.inner_product(vector, vector, base_point)
        return gs.real(sq_norm)

    def random_unit_tangent_vec(self, base_point=None, n_vectors=1):
        """Generate a random unit tangent vector at a given point.

        Parameters
        ----------
        base_point : array-like, shape=[..., dim]
            Point.
        n_vectors : float
            Number of vectors to be generated at base_point.
            For vectorization purposes n_vectors can be greater than 1 iff
            base_point consists of a single point.

        Returns
        -------
        normalized_vector : array-like, shape=[..., n_vectors, dim]
            Random unit tangent vector at base_point.
        """
        is_batch = check_is_batch(self._space.point_ndim, base_point)
        if is_batch and n_vectors > 1:
            raise ValueError(
                "Several tangent vectors is only applicable to a single base point."
            )
        dtype = gs.get_default_cdtype() if base_point is None else base_point.dtype
        point_shape = self._space.shape
        random_vector = gs.squeeze(
            gs.cast(gs.random.rand(n_vectors, *point_shape), dtype=dtype)
            + 1j * gs.cast(gs.random.rand(n_vectors, *point_shape), dtype=dtype)
        )
        normalized_vector = self.normalize(random_vector, base_point)
        return gs.squeeze(normalized_vector)



import itertools
import random

import pytest

import geomstats.backend as gs
from geomstats.test.random import RandomDataGenerator
from geomstats.test.test_case import TestCase
from geomstats.test.vectorization import generate_vectorization_data


class PermutationGroup:
    """Permutation group.

    A basic (non-optimized) implementation of the permutation group with points
    represented as 1d arrays for testing purposes.
    """

    def __init__(self, dim):
        self.dim = dim
        self.shape = (dim,)
        self.identity = gs.arange(dim)

        self._all_perms = list(itertools.permutations(range(self.dim), self.dim))

    def random_point(self, n_samples=1):
        """Sample random point.

        Parameters
        ----------
        n_samples : int

        Returns
        -------
        random_point : array-like, shape=[..., dim]
        """
        random_point = gs.array(random.sample(self._all_perms, n_samples))
        if n_samples == 1:
            return random_point[0]

        return random_point

    def _inverse_single(self, group_elem):
        """Inverse element.

        Parameters
        ----------
        group_elem : array-like, shape=[dim]

        Returns
        -------
        inverse_group_elem : array-like, shape=[dim]
        """
        inverse_group_elem = gs.zeros_like(group_elem)
        for val, index in enumerate(group_elem):
            inverse_group_elem[index] = val
        return inverse_group_elem

    def inverse(self, group_elem):
        """Inverse element.

        Parameters
        ----------
        group_elem : array-like, shape=[..., dim]

        Returns
        -------
        inverse_group_elem : array-like, shape=[..., dim]
        """
        if group_elem.ndim == 1:
            return self._inverse_single(group_elem)

        return gs.stack(
            [self._inverse_single(group_elem_) for group_elem_ in group_elem]
        )


class GroupActionTestCase(TestCase):
    """Group action test case.

    Requires: `group_action`, `space`, `group`.
    """

    def setup_method(self):
        if not hasattr(self, "data_generator"):
            self.data_generator = RandomDataGenerator(self.space)
            self.group_data_generator = RandomDataGenerator(self.group)

    @pytest.mark.random
    def test_identity_action(self, n_points, atol):
        point = self.data_generator.random_point(n_points)
        orbit_point = self.group_action(self.group.identity, point)
        self.assertAllClose(point, orbit_point)

    def test_action(self, group_elem, point, expected, atol):
        orbit_point = self.group_action(group_elem, point)
        self.assertAllClose(orbit_point, expected, atol=atol)

    @pytest.mark.vec
    def test_action_vec(self, n_reps, atol):
        group_elem = self.group_data_generator.random_point()
        point = self.data_generator.random_point()

        expected = self.group_action(group_elem, point)

        vec_data = generate_vectorization_data(
            data=[
                dict(group_elem=group_elem, point=point, expected=expected, atol=atol)
            ],
            arg_names=["group_elem", "point"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_inverse_action(self, n_points, atol):
        group_elem = self.group_data_generator.random_point(n_points)
        point = self.data_generator.random_point(n_points)

        orbit_point = self.group_action(group_elem, point)
        point_ = self.group_action(self.group.inverse(group_elem), orbit_point)
        self.assertAllClose(point_, point, atol)



"""Open hemisphere.

For more details, check section 7.4.1 of [T2022]_.

Lead author: Olivier Bisson.

References
----------
.. [T2022] Yann Thanwerdas. Riemannian and stratified
    geometries on covariance and correlation matrices. Differential
    Geometry [math.DG]. Université Côte d'Azur, 2022.
"""

import geomstats.backend as gs
from geomstats.geometry.base import OpenSet
from geomstats.geometry.diffeo import InvolutionDiffeomorphism
from geomstats.geometry.hyperboloid import Hyperboloid
from geomstats.geometry.hypersphere import Hypersphere
from geomstats.geometry.product_manifold import ProductManifold, ProductRiemannianMetric
from geomstats.geometry.pullback_metric import PullbackDiffeoMetric


class OpenHemisphereToHyperboloidDiffeo(InvolutionDiffeomorphism):
    """A diffeomorphism between the open hemisphere and the hyperboloid."""

    def __call__(self, base_point):
        """Diffeomorphism at base point."""
        return_point = gs.copy(base_point)
        return_point[..., 0] = 1.0
        first_term = base_point[..., 0]
        return gs.einsum("...,...i->...i", 1 / first_term, return_point)

    def tangent(self, tangent_vec, base_point=None, image_point=None):
        """Tangent diffeomorphism at base point."""
        if base_point is None:
            base_point = self.inverse(image_point)

        coeffs = tangent_vec[..., 0] / base_point[..., 0]
        image_tangent_vec_0 = gs.array(-tangent_vec[..., 0] / base_point[..., 0])
        image_tangent_vec_other = tangent_vec[..., 1:] - gs.einsum(
            "...,...i->...i", coeffs, base_point[..., 1:]
        )

        image_tangent_vec = gs.concatenate(
            [gs.expand_dims(image_tangent_vec_0, axis=-1), image_tangent_vec_other],
            axis=-1,
        )
        return gs.einsum("...,...i->...i", 1 / base_point[..., 0], image_tangent_vec)


class OpenHemisphere(OpenSet):
    r"""Open hemisphere.

    An open set of the hypersphere where the first coordinate is always
    positive.

    .. math::

        \mathrm{HS}^{k-1}=\left\{x \in \mathbb{R}^k: \|x\|=1
        \text { and } x_0>0\right\}
    """

    def __init__(self, dim, equip=True):
        self.dim = dim
        super().__init__(
            dim=dim,
            intrinsic=False,
            embedding_space=Hypersphere(dim, equip=equip),
            equip=equip,
        )

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return OpenHemispherePullbackMetric

    def belongs(self, point, atol=gs.atol):
        """Check if a point belongs to the open hemisphere."""
        is_on_sphere = self.embedding_space.belongs(point)
        is_on_upper_part = gs.greater(point[..., 0], 0.0)
        return gs.logical_and(is_on_sphere, is_on_upper_part)

    def projection(self, point):
        """Project a point on the open hemisphere.

        Parameters
        ----------
        point : array-like, shape=[..., dim]
            Point in embedding hypersphere space.

        Returns
        -------
        projected_point : array-like, shape=[..., dim]
            Point projected on the open hemisphere.
        """
        proj_point = self.embedding_space.projection(point)
        proj_point[..., 0] = gs.abs(proj_point[..., 0])
        return proj_point


class OpenHemispherePullbackMetric(PullbackDiffeoMetric):
    """Pullback diffeo metric for Open Hemisphere.

    Pulls back metric from hyperboloid.
    """

    def __init__(self, space):
        image_space = Hyperboloid(dim=space.dim)
        diffeo = OpenHemisphereToHyperboloidDiffeo()
        super().__init__(space=space, diffeo=diffeo, image_space=image_space)


class OpenHemispheresProduct(ProductManifold):
    r"""A consecutively factor-dim increasing product manifold of open hemispheres.

    .. math::

        HS^1\times \dots \times HS(n)
    """

    def __init__(self, n, equip=True):
        factors = [OpenHemisphere(dim=dim, equip=True) for dim in range(1, n)]

        super().__init__(
            factors=factors,
            point_ndim=1,
            equip=equip,
        )

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return OpenHemispheresProductMetric


class OpenHemispheresProductMetric(ProductRiemannianMetric):
    """Define the product metric on these manifolds."""
